@@ -1,25 +1,24 @@
 /* This file is part of GNU RADIUS.
- * Copyright (C) 2000,2001, Sergey Poznyakoff
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
- */
-/* $Id: acconfig.h,v 1.6 2001/04/26 11:25:00 gray Exp $ */
+   Copyright (C) 2000,2001, Sergey Poznyakoff
+ 
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+ 
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+ 
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation, 
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 
-#undef PW_AUTH_UDP_PORT
-#undef PW_ACCT_UDP_PORT
+/* $Id: acconfig.h,v 1.16 2001/12/20 16:53:56 gray Exp $ */
+
+#undef DEF_AUTH_PORT
+#undef DEF_ACCT_PORT
 
 /* Target RADIUS was built for */
 #undef BUILD_TARGET 
@@ -30,8 +29,9 @@
 /* Directory for pidfile */
 #undef RADPID_DIR
 
-#define SQL_MYSQL    1
-#define SQL_POSTGRES 2
+#define DBM_DBM  1
+#define DBM_NDBM 2
+
 
 @TOP@
 
@@ -57,7 +57,7 @@
 #undef u_long
 /* ************************************************************************** */
 
-#undef HAVE_GETOPT_LONG
+#undef HAVE_GNU_GETOPT
 
 #undef HAVE_SYS_ERRLIST
 
@@ -71,6 +71,8 @@
 
 #undef HAVE_LIBPQ
 
+#undef HAVE_LIBODBC
+
 /* Define this to disable shadow support */
 #undef PWD_SHADOW
 
@@ -81,6 +83,9 @@
 /* alwys be denied access */
 #undef DENY_SHELL
 
+/* Define this to enable Guile interface */
+#undef USE_SERVER_GUILE
+
 /* Define this to enable SNMP support */
 #undef USE_SNMP
 
@@ -90,12 +95,15 @@
 /* Define this to enable MYSQL subsystem of SQL support */
 #undef USE_SQL_MYSQL
 
+/* Define this to enable POSTGRES subsystem of SQL support */
+#undef USE_SQL_POSTGRES
+
+/* Define this to enable ODBC subsystem of SQL support */
+#undef USE_SQL_ODBC
+
 /* Define this to enable Livingston-compatible menus */
 #undef USE_LIVINGSTON_MENUS
 
-/* Define this to enable notification */
-#undef USE_NOTIFY
-
 /* Port to use to connect to MYSQL */
 /* leave undefined to use default one */
 #undef RAD_MYSQL_PORT
@@ -106,10 +114,6 @@
 /* Define this if you wish the DBM support */
 #undef USE_DBM
 
-#undef DBM
-
-#undef NDBM
-
 #undef YACC_DEBUG
 
 #if defined(sun) && !defined(__EXTENSIONS__)
@@ -133,3 +137,7 @@
 #  define __extension__(c) 1
 # endif
 #endif
+
+#ifdef RADIUS_DEBUG
+# include <debugmod.h>
+#endif
