@@ -1,26 +1,24 @@
 /* This file is part of GNU RADIUS.
- * Copyright (C) 2000,2001, Sergey Poznyakoff
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
- */
+   Copyright (C) 2000,2001, Sergey Poznyakoff
+ 
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+ 
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+ 
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation,
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 
-#define RADIUS_MODULE 6
+#define RADIUS_MODULE_AUTH_C
 #ifndef lint
 static char rcsid[] =
-"@(#) $Id: auth.c,v 1.18 2001/04/17 06:43:34 gray Exp $";
+"@(#) $Id: auth.c,v 1.47 2001/12/21 14:08:18 gray Exp $";
 #endif
 
 #ifdef HAVE_CONFIG_H
@@ -32,6 +30,7 @@ static char rcsid[] =
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <ctype.h>
 #include <pwd.h>
 #include <time.h>
 #include <unistd.h>
@@ -53,19 +52,24 @@ static char rcsid[] =
 #endif
 
 #include <radiusd.h>
+#include <rewrite.h>
 #if defined(USE_SQL)
 # include <radsql.h>
 #endif
 #include <timestr.h>
+#include <envar.h>
+#include <obstack1.h>
 
 #if !defined(__linux__) && !defined(__GLIBC__)
   extern char *crypt();
 #endif
 
 static int pw_expired(UINT4 exptime);
-static int check_disable(char *username, char **user_msg);
-static int check_expiration(VALUE_PAIR *check_item, char **user_msg);
 static int unix_pass(char *name, char *passwd);
+static int rad_check_password(RADIUS_REQ *radreq, 
+			      VALUE_PAIR *check_item, VALUE_PAIR *namepair,
+			      char **user_msg,
+			      char *userpass);
 
 /*
  * Tests to see if the users password has expired.
@@ -97,62 +101,22 @@ pw_expired(exptime)
 	return 0;
 }
 
+char *username_valid_chars;
+
 /*
  * Check if the username is valid. Valid usernames consist of 
- * alphanumeric characters and symbols from valid_char[]
+ * alphanumeric characters and symbols from username_valid_chars[]
  * array
  */
 int
 check_user_name(p)
 	char *p;
 {
-	static char valid_char[] = ".-_!@#$%^&";
-	for (; *p && (isalnum(*p) || strchr(valid_char, *p)); p++)
+	for (; *p && (isalnum(*p) || strchr(username_valid_chars, *p)); p++)
 		;
 	return *p;
 }
 
-int
-check_disable(username, user_msg)
-	char *username;
-	char **user_msg;
-{
-	if (get_deny(username)) {
-		*user_msg = make_string(
-			    _("Sorry, your account is currently closed\r\n"));
-		return -1;
-	}
-	return 0;
-}
-
-/*
- * Check if account has expired, and if user may login now.
- */
-int
-check_expiration(check_item, user_msg)
-	VALUE_PAIR *check_item;
-	char **user_msg;
-{
-	int result, rc;
-	VALUE_PAIR *pair;
-	char umsg[80];
-	
-	result = AUTH_OK;
-	if (pair = pairfind(check_item, DA_EXPIRATION)) {
-		rc = pw_expired(pair->lvalue);
-		if (rc < 0) {
-			result = AUTH_FAIL;
-			*user_msg = make_string(_("Password Has Expired\r\n"));
-		} else if (rc > 0) {
-			radsprintf(umsg, sizeof(umsg),
-				   _("Password Will Expire in %d Days\r\n"),
-				   rc);
-			*user_msg = make_string(umsg);
-		}
-	}
-
-	return result;
-}
 /*
  * Check the users password against UNIX password database.
  */
@@ -204,7 +168,7 @@ unix_pass(name, passwd)
 	 * Users with a certain shell are always denied access.
 	 */
 	if (strcmp(pwd->pw_shell, DENY_SHELL) == 0) {
-		radlog(L_AUTH, _("unix_pass: [%s]: invalid shell"), name);
+		radlog(L_NOTICE, _("unix_pass: [%s]: invalid shell"), name);
 		return -1;
 	}
 #endif
@@ -215,7 +179,7 @@ unix_pass(name, passwd)
 	 */
 	if (spwd && spwd->sp_expire > 0 &&
 	    (time(NULL) / 86400) > spwd->sp_expire) {
-		radlog(L_AUTH, _("unix_pass: [%s]: password has expired"), name);
+		radlog(L_NOTICE, _("unix_pass: [%s]: password has expired"), name);
 		return -1;
 	}
 #endif
@@ -225,7 +189,7 @@ unix_pass(name, passwd)
 	 * Check if the account is locked.
 	 */
 	if (pr_pw->uflg.fg_lock != 1) {
-		radlog(L_AUTH, _("unix_pass: [%s]: account locked"), name);
+		radlog(L_NOTICE, _("unix_pass: [%s]: account locked"), name);
 		return -1;
 	}
 #endif /* OSFC2 */
@@ -256,215 +220,192 @@ unix_pass(name, passwd)
  *			AUTH_REJECT  Rejected
  *			
  */
-/*ARGSUSED*/
 int
-rad_check_password(authreq, activefd, check_item, namepair,
-		   pw_digest, user_msg, userpass)
-	AUTH_REQ   *authreq;
-	int        activefd;
+rad_check_password(radreq, check_item, namepair, user_msg, userpass)
+	RADIUS_REQ   *radreq;
 	VALUE_PAIR *check_item;
 	VALUE_PAIR *namepair;
-	u_char     *pw_digest;
 	char       **user_msg;
 	char       *userpass;
 {
-	char		string[AUTH_STRING_LEN];
-	char		*ptr;
-	char		name[AUTH_STRING_LEN];
-	VALUE_PAIR	*auth_type_pair;
-	VALUE_PAIR	*password_pair;
-	VALUE_PAIR	*auth_item;
-	VALUE_PAIR	*tmp;
-	int		auth_type = -1;
-	int		i;
-	int		result;
-        VALUE_PAIR      *p;
-	char            *authdata = NULL;
+	char *ptr;
+	char *real_password = NULL;
+	char name[AUTH_STRING_LEN];
+	VALUE_PAIR *auth_item;
+	VALUE_PAIR *tmp;
+	int auth_type = -1;
+	int i;
+	int length;
+	int result;
+	char *authdata = NULL;
+	char pw_digest[AUTH_VECTOR_LEN];
 	
 	result = AUTH_OK;
 	userpass[0] = 0;
-	string[0] = 0;
 
-	/*
-	 * Look for matching check items. We skip the whole lot
-	 * if the authentication type is DV_AUTH_TYPE_ACCEPT or
-	 * DV_AUTH_TYPE_REJECT.
-	 */
-	if ((auth_type_pair = pairfind(check_item, DA_AUTH_TYPE)) != NULL)
-		auth_type = auth_type_pair->lvalue;
+	/* Process immediate authentication types */
+	if ((tmp = avl_find(check_item, DA_AUTH_TYPE)) != NULL)
+		auth_type = tmp->lvalue;
 
 	if (auth_type == DV_AUTH_TYPE_ACCEPT)
 		return AUTH_OK;
 
-	if (auth_type == DV_AUTH_TYPE_REJECT) {
-		*user_msg = NULL;
+	if (auth_type == DV_AUTH_TYPE_REJECT) 
 		return AUTH_REJECT;
-	}
 
-        if ((p = pairfind(check_item, DA_AUTH_DATA)) != NULL) {
-		authdata = p->strvalue;
-        }
+	/* Find the password sent by the user. If it's not present,
+	   authentication fails. */
 	
-	/*
-	 * Find the password sent by the user. It SHOULD be there,
-	 * if it's not authentication fails.
-	 */
-	if ((auth_item = pairfind(authreq->request, DA_CHAP_PASSWORD)) == NULL)
-		auth_item = pairfind(authreq->request, DA_PASSWORD);
+	if (!(auth_item = avl_find(radreq->request, DA_CHAP_PASSWORD))) {
+		auth_item = avl_find(radreq->request, DA_PASSWORD);
+		/* Decrypt the password. */
+		if (auth_item) {
+			if (auth_item->strlength == 0)
+				userpass[0] = 0;
+			else
+				req_decrypt_password(userpass, radreq,
+						     auth_item);
+		}
+	}
+		
 	if (auth_item == NULL)
 		return AUTH_FAIL;
+	
+	/* Set up authentication data */
+	if ((tmp = avl_find(check_item, DA_AUTH_DATA)) != NULL) 
+		authdata = tmp->strvalue;
 
-	/*
-	 * Find the password from the users file.
-	 */
-	if ((password_pair = pairfind(check_item, DA_CRYPT_PASSWORD)) != NULL)
-		auth_type = DV_AUTH_TYPE_CRYPT_LOCAL;
-	else
-		password_pair = pairfind(check_item, DA_PASSWORD);
+	/* Find the 'real' password */
+	tmp = avl_find(check_item, DA_PASSWORD);
+	if (tmp)
+		real_password = estrdup(tmp->strvalue);
+	else if (tmp = avl_find(check_item, DA_PASSWORD_LOCATION)) {
+		switch (tmp->lvalue) {
+		case DV_PASSWORD_LOCATION_SQL:
+#ifdef USE_SQL
+			real_password = rad_sql_pass(radreq, authdata);
+			if (!real_password)
+				return AUTH_NOUSER;
+#else
+			radlog(L_ERR,
+			       _("%s: SQL authentication not available"),
+			       namepair->strvalue);
+			return AUTH_NOUSER;
+#endif
+			break;
+		/*NOTE: add any new location types here */
+		default:
+			radlog(L_ERR,
+			       _("unknown Password-Location value: %d"),
+			       tmp->lvalue);
+			return AUTH_FAIL;
+		}
+	}
 
-	/*
-	 * See if there was a Prefix or Suffix included.
-	 */
+	/* Process any prefixes/suffixes. */
 	strip_username(1, namepair->strvalue, check_item, name);
 
-	/*
-	 * Decrypt the password.
-	 */
-	if (auth_item != NULL && auth_item->attribute == DA_PASSWORD) {
-		if (auth_item->strlength == 0)
-			string[0] = 0;
-		else {
-			memcpy(string, auth_item->strvalue, AUTH_PASS_LEN);
-			for (i = 0;i < AUTH_PASS_LEN;i++) {
-				string[i] ^= pw_digest[i];
-			}
-			string[AUTH_PASS_LEN] = '\0';
-		}
-		strcpy(userpass, string);
-	}
-
 	debug(1,
-		("auth_type=%d, string=%s, namepair=%s, password_pair=%s",
-		 auth_type, string, name,
-		 password_pair ? password_pair->strvalue : ""));
-
-	if ((p = pairfind(check_item, DA_AUTH_DATA)) != NULL) 
-		authdata = p->strvalue;
+		("auth_type=%d, userpass=%s, name=%s, password=%s",
+		 auth_type, userpass, name,
+		 real_password ? real_password : "NONE"));
 
 	switch (auth_type) {
-		case DV_AUTH_TYPE_SYSTEM:
-			debug(1, ("  auth: System"));
-			/*
-			 * Check the password against /etc/passwd.
-			 */
-			if (unix_pass(name, string) != 0)
-				result = AUTH_FAIL;
-			break;
-		case DV_AUTH_TYPE_PAM:
+	case DV_AUTH_TYPE_SYSTEM:
+		debug(1, ("  auth: System"));
+		if (unix_pass(name, userpass) != 0)
+			result = AUTH_FAIL;
+		break;
+	case DV_AUTH_TYPE_PAM:
 #ifdef USE_PAM
-			debug(1, ("  auth: Pam"));
-			/*
-			 * Provide defaults for authdata
-			 */
-			if (authdata == NULL &&
-			    (p = pairfind(check_item, DA_PAM_AUTH)) != NULL) {
-				authdata = p->strvalue;
-			}
-			authdata = authdata ? authdata : PAM_DEFAULT_TYPE;
-			if (pam_pass(name, string, authdata, user_msg) != 0)
-				result = AUTH_FAIL;
+		debug(1, ("  auth: Pam"));
+		/* Provide defaults for authdata */
+		if (authdata == NULL &&
+		    (tmp = avl_find(check_item, DA_PAM_AUTH)) != NULL) {
+			authdata = tmp->strvalue;
+		}
+		authdata = authdata ? authdata : PAM_DEFAULT_TYPE;
+		if (pam_pass(name, userpass, authdata, user_msg) != 0)
+			result = AUTH_FAIL;
 #else
-			radlog(L_ERR,
-			       _("%s: PAM authentication not available"),
-			       name);
-			result = AUTH_NOUSER;
+		radlog(L_ERR,
+		       _("%s: PAM authentication not available"),
+		       name);
+		result = AUTH_NOUSER;
 #endif
-			break;
-		case DV_AUTH_TYPE_MYSQL:
-#ifdef USE_SQL
-			result = rad_sql_pass(authreq, authdata, string);
-#else
-			radlog(L_ERR,
-			       _("%s: MYSQL authentication not available"),
-			       name);
-			result = AUTH_NOUSER;
-#endif
-			break;
-		case DV_AUTH_TYPE_CRYPT_LOCAL:
-			debug(1, ("  auth: Crypt"));
-			if (password_pair == NULL) {
-				result = string[0] ? AUTH_FAIL : AUTH_OK;
-				break;
-			}
-			if (strcmp(password_pair->strvalue,
-			    md5crypt(string, password_pair->strvalue)) != 0)
-				result = AUTH_FAIL;
-			break;
-		case DV_AUTH_TYPE_LOCAL:
-			debug(1, ("  auth: Local"));
-			/*
-			 * Local password is just plain text.
-	 		 */
-			if (auth_item->attribute != DA_CHAP_PASSWORD) {
-				/*
-				 *	Plain text password.
-				 */
-				if (password_pair == NULL ||
-				    strcmp(password_pair->strvalue, string)!=0)
-					result = AUTH_FAIL;
-				break;
-			}
+		break;
 
-			/*
-			 * CHAP - calculate MD5 sum over CHAP-ID,
-			 * plain-text password and the Chap-Challenge.
-			 * Compare to Chap-Response (strvalue + 1).
-			 *
-			 * FIXME: might not work with Ascend because
-			 * we use vp->strlength, and Ascend gear likes
-			 * to send an extra '\0' in the string!
-			 */
-			strcpy(string, "{chap-password}");
-			if (password_pair == NULL) {
-				result= AUTH_FAIL;
-				break;
-			}
-			i = 0;
-			ptr = string;
-			*ptr++ = *auth_item->strvalue;
-			i++;
-			memcpy(ptr, password_pair->strvalue,
-				password_pair->strlength);
-			ptr += password_pair->strlength;
-			i += password_pair->strlength;
-			/*
-			 * Use Chap-Challenge pair if present,
-			 * Request-Authenticator otherwise.
-			 */
-			if ((tmp = pairfind(authreq->request,
-					    DA_CHAP_CHALLENGE)) != NULL) {
-				memcpy(ptr, tmp->strvalue, tmp->strlength);
-				i += tmp->strlength;
-			} else {
-				memcpy(ptr, authreq->vector, AUTH_VECTOR_LEN);
-				i += AUTH_VECTOR_LEN;
-			}
-			md5_calc(pw_digest, (u_char*) string, i);
-
-			/*
-			 * Compare them
-			 */
-			if (memcmp(pw_digest, auth_item->strvalue + 1,
-					CHAP_VALUE_LENGTH) != 0)
-				result = AUTH_FAIL;
-			else
-				strcpy(userpass, password_pair->strvalue);
-			break;
-		default:
+	case DV_AUTH_TYPE_CRYPT_LOCAL:
+		debug(1, ("  auth: Crypt"));
+		if (real_password == NULL) {
 			result = AUTH_FAIL;
 			break;
+		}
+		if (strcmp(real_password,
+			   md5crypt(userpass, real_password)) != 0)
+			result = AUTH_FAIL;
+		break;
+		
+	case DV_AUTH_TYPE_LOCAL:
+		debug(1, ("  auth: Local"));
+		/* Local password is just plain text. */
+		if (auth_item->attribute != DA_CHAP_PASSWORD) {
+			if (real_password == NULL ||
+			    strcmp(real_password, userpass) != 0)
+				result = AUTH_FAIL;
+			break;
+		}
+
+		/* CHAP - calculate MD5 sum over CHAP-ID,
+		   plain-text password and the Chap-Challenge.
+		   Compare to Chap-Response (strvalue + 1).
+		 
+		   FIXME: might not work with Ascend because
+		   we use vp->strlength, and Ascend gear likes
+		   to send an extra '\0' in the string! */
+		
+		strcpy(userpass, "{chap-password}");
+		if (real_password == NULL) {
+			result = AUTH_FAIL;
+			break;
+		}
+		i = 0;
+		ptr = userpass;
+		*ptr++ = *auth_item->strvalue;
+		i++;
+		length = strlen(real_password);
+		memcpy(ptr, real_password, length);
+		ptr += length;
+		i += length;
+		
+		/* Use Chap-Challenge pair if present,
+		   Request-Authenticator otherwise. */
+		if (tmp = avl_find(radreq->request, DA_CHAP_CHALLENGE)) {
+			memcpy(ptr, tmp->strvalue, tmp->strlength);
+			i += tmp->strlength;
+		} else {
+			memcpy(ptr, radreq->vector, AUTH_VECTOR_LEN);
+			i += AUTH_VECTOR_LEN;
+		}
+		md5_calc(pw_digest, (u_char*) userpass, i);
+		
+		/* Compare them */
+		if (memcmp(pw_digest, auth_item->strvalue + 1,
+			   CHAP_VALUE_LENGTH) != 0)
+			result = AUTH_FAIL;
+		else
+			strcpy(userpass, real_password);
+		break;
+	default:
+		result = AUTH_FAIL;
+		break;
 	}
 
+	if (real_password) {
+		/* just in case: */
+		memset(real_password, 0, strlen(real_password)); 
+		efree(real_password);
+	}
 	return result;
 }
 
@@ -474,8 +415,8 @@ rad_check_password(authreq, activefd, check_item, namepair,
  *	process the hints and huntgroups file.
  */
 int
-rad_auth_init(authreq, activefd)
-	AUTH_REQ *authreq;
+rad_auth_init(radreq, activefd)
+	RADIUS_REQ *radreq;
 	int       activefd;
 {
 	VALUE_PAIR	*namepair;
@@ -486,62 +427,40 @@ rad_auth_init(authreq, activefd)
 	/*
 	 * Get the username from the request
 	 */
-	namepair = pairfind(authreq->request, DA_USER_NAME);
+	namepair = avl_find(radreq->request, DA_USER_NAME);
 
 	if ((namepair == (VALUE_PAIR *)NULL) || 
 	   (strlen(namepair->strvalue) <= 0)) {
 		radlog(L_ERR, _("No username: [] (from nas %s)"),
-		       nas_name2(authreq));
-		stat_inc(auth, authreq->ipaddr, num_bad_req);
-		authfree(authreq);
+		       nas_request_to_name(radreq));
+		stat_inc(auth, radreq->ipaddr, num_bad_req);
 		return -1;
 	}
 	debug(1,("checking username: %s", namepair->strvalue));
 	if (check_user_name(namepair->strvalue)) {
-		radlog(L_AUTH, _("Malformed username: [%s] (from nas %s)"),
+		radlog(L_ERR, _("Malformed username: [%s] (from nas %s)"),
 		       namepair->strvalue,
-		       nas_name2(authreq));
-		stat_inc(auth, authreq->ipaddr, num_bad_req);
-		authfree(authreq);
+		       nas_request_to_name(radreq));
+		stat_inc(auth, radreq->ipaddr, num_bad_req);
 		return -1;
 	}
 		
-	/*
-	 * Verify the client and Calculate the MD5 Password Digest
-	 */
-	if (calc_digest(authreq->digest, authreq) != 0) {
-		/*
-		 * We don't respond when this fails
-		 */
-		radlog(L_NOTICE,
-		       _("from client %s - Security Breach: %s"),
-		       client_name(authreq->ipaddr), namepair->strvalue);
-		stat_inc(auth, authreq->ipaddr, num_bad_auth);
-		authfree(authreq);
-		return -1;
-	}
-
-#ifdef USE_SQL
-	if (p = create_pair(DA_QUEUE_ID, 0, NULL, (qid_t)authreq))
-		pairadd(&authreq->request, p);
-#endif
 	/*
 	 * Add any specific attributes for this username.
 	 */
-	hints_setup(authreq->request);
+	hints_setup(radreq);
 
 	if (auth_detail)
-		write_detail(authreq, -1, "detail.auth");
+		write_detail(radreq, -1, "detail.auth");
 
 	/*
 	 * See if the user has access to this huntgroup.
 	 */
-	if (!huntgroup_access(authreq)) {
-		radlog(L_AUTH, _("No huntgroup access: [%s] (from nas %s)"),
-			namepair->strvalue, nas_name2(authreq));
-		rad_send_reply(PW_AUTHENTICATION_REJECT, authreq,
-			       authreq->request, NULL, activefd);
-		authfree(authreq);
+	if (!huntgroup_access(radreq)) {
+		radlog(L_NOTICE, _("No huntgroup access: [%s] (from nas %s)"),
+			namepair->strvalue, nas_request_to_name(radreq));
+		rad_send_reply(RT_AUTHENTICATION_REJECT, radreq,
+			       radreq->request, NULL, activefd);
 		return -1;
 	}
 
@@ -561,7 +480,8 @@ enum auth_state {
 	as_realmuse,
 	as_simuse, 
 	as_time, 
-	as_ttl, 
+	as_eval,
+	as_scheme,
 	as_ipaddr, 
 	as_exec_wait, 
 	as_cleanup_cbkid, 
@@ -581,7 +501,7 @@ enum list_id {
 };
 
 typedef struct auth_mach {
-	AUTH_REQ   *req;
+	RADIUS_REQ *req;
 	VALUE_PAIR *user_check;
 	VALUE_PAIR *user_reply;
 	VALUE_PAIR *proxy_pairs;
@@ -590,8 +510,10 @@ typedef struct auth_mach {
 	VALUE_PAIR *namepair;
 	VALUE_PAIR *check_pair;
 	VALUE_PAIR *timeout_pair;
-	char       userpass[AUTH_STRING_LEN];
+	char       userpass[AUTH_STRING_LEN+1];
+
 	char       *user_msg;
+	struct obstack msg_stack;
 	
 	char       *clid;
 	enum auth_state state;
@@ -599,13 +521,14 @@ typedef struct auth_mach {
 
 static void sfn_init(MACH*);
 static void sfn_validate(MACH*);
+static void sfn_eval_reply(MACH*);
+static void sfn_scheme(MACH*);
 static void sfn_service(MACH*);
 static void sfn_disable(MACH*);
 static void sfn_service_type(MACH*);
 static void sfn_realmuse(MACH*);
 static void sfn_simuse(MACH*);
 static void sfn_time(MACH*);
-static void sfn_ttl(MACH*);
 static void sfn_ipaddr(MACH*);
 static void sfn_exec_wait(MACH*);
 static void sfn_cleanup_cbkid(MACH*);
@@ -614,6 +537,7 @@ static void sfn_ack(MACH*);
 static void sfn_exec_nowait(MACH*);
 static void sfn_reject(MACH*);
 static VALUE_PAIR *timeout_pair(MACH *m);
+static int check_expiration(MACH *m);
 
 
 struct auth_state_s {
@@ -646,11 +570,14 @@ struct auth_state_s states[] = {
 	as_simuse,       as_time,
 	                 DA_SIMULTANEOUS_USE, L_check, sfn_simuse,
 	
-	as_time,         as_ttl,
+	as_time,         as_eval,
 	                 DA_LOGIN_TIME,   L_check, sfn_time,
 	
-	as_ttl,          as_ipaddr,
-	                 0,               L_null, sfn_ttl,
+	as_eval,         as_scheme,
+	                 0,               L_null,  sfn_eval_reply,
+
+	as_scheme,       as_ipaddr,
+	                 DA_SCHEME_PROCEDURE, L_reply, sfn_scheme,
 	
 	as_ipaddr,       as_exec_wait,
 	                 0,               L_null, sfn_ipaddr,
@@ -680,6 +607,8 @@ struct auth_state_s states[] = {
 static void auth_log(MACH *m, char *diag, char *pass, char *reason,
 		     char *addstr);
 static int is_log_mode(MACH *m, int mask);
+static void auth_format_msg(MACH *m, int msg_id);
+static char *auth_finish_msg(MACH *m);
 
 void
 auth_log(m, diag, pass, reason, addstr)
@@ -690,7 +619,7 @@ auth_log(m, diag, pass, reason, addstr)
 	char *addstr;
 {
 	if (reason)
-		radlog(L_AUTH,
+		radlog(L_NOTICE,
 		       _("%s: [%s%s%s]: %s%s: CLID %s (from nas %s)"),
 		       diag,
 		       m->namepair->strvalue,
@@ -699,16 +628,16 @@ auth_log(m, diag, pass, reason, addstr)
 		       reason,
 		       addstr ? addstr : "",
 		       m->clid,
-		       nas_name2(m->req));
+		       nas_request_to_name(m->req));
 	else
-		radlog(L_AUTH,
+		radlog(L_NOTICE,
 		       _("%s: [%s%s%s]: CLID %s (from nas %s)"),
 		       diag,
 		       m->namepair->strvalue,
 		       pass ? "/" : "",
 		       pass ? pass : "",
 		       m->clid,
-		       nas_name2(m->req));
+		       nas_request_to_name(m->req));
 }
 
 int
@@ -716,45 +645,99 @@ is_log_mode(m, mask)
 	MACH *m;
 	int mask;
 {
-	int mode = log_mode;
+        int mode = log_mode;
+        int xmask = 0;
 #ifdef DA_LOG_MODE_MASK
-	VALUE_PAIR *p;
-	
-	for (p = pairfind(m->user_check, DA_LOG_MODE_MASK);
-	     p;
-	     p = p->next ? pairfind(p->next, DA_LOG_MODE_MASK) : NULL) 
-		mode &= ~p->lvalue;
-	for (p = pairfind(m->req->request, DA_LOG_MODE_MASK);
-	     p;
-	     p = p->next ? pairfind(p->next, DA_LOG_MODE_MASK) : NULL) 
-		mode &= ~p->lvalue;
+        VALUE_PAIR *p;
+
+        for (p = avl_find(m->user_check, DA_LOG_MODE_MASK);
+             p;
+             p = p->next ? avl_find(p->next, DA_LOG_MODE_MASK) : NULL)
+                xmask |= p->lvalue;
+        for (p = avl_find(m->req->request, DA_LOG_MODE_MASK);
+             p;
+             p = p->next ? avl_find(p->next, DA_LOG_MODE_MASK) : NULL)
+                xmask |= p->lvalue;
 #endif
-	return mode & mask;
+        return (mode & ~xmask) & mask;
 }
 
+void
+auth_format_msg(m, msg_id)
+	MACH *m;
+	int msg_id;
+{
+	int len = strlen(message_text[msg_id]);
+	obstack_grow(&m->msg_stack, message_text[msg_id], len);
+}
+
+char *
+auth_finish_msg(m)
+	MACH *m;
+{
+	if (m->user_msg)
+		obstack_grow(&m->msg_stack, m->user_msg, strlen(m->user_msg));
+	obstack_1grow(&m->msg_stack, 0);
+	return radius_xlate(&m->msg_stack, obstack_finish(&m->msg_stack),
+			    m->req, m->user_reply);
+}
+
+
+/*
+ * Check if account has expired, and if user may login now.
+ */
 int
-rad_authenticate(authreq, activefd)
-	AUTH_REQ  *authreq;
+check_expiration(m)
+	MACH *m;
+{
+	int result, rc;
+	VALUE_PAIR *pair;
+	
+	result = AUTH_OK;
+	if (pair = avl_find(m->user_check, DA_EXPIRATION)) {
+		rc = pw_expired(pair->lvalue);
+		if (rc < 0) {
+			result = AUTH_FAIL;
+			auth_format_msg(m, MSG_PASSWORD_EXPIRED);
+		} else if (rc > 0) {
+			VALUE_PAIR *pair;
+			pair = avp_create(DA_PASSWORD_EXPIRE_DAYS, 0, NULL,
+					  rc/86400);
+			avl_add_pair(&m->user_reply, pair);
+			auth_format_msg(m, MSG_PASSWORD_EXPIRE_WARNING);
+		}
+	}
+
+	return result;
+}
+
+
+int
+rad_authenticate(radreq, activefd)
+	RADIUS_REQ  *radreq;
 	int        activefd;
 {
 	enum auth_state oldstate;
 	struct auth_state_s *sp;
 	struct auth_mach m;
-
 #ifdef USE_LIVINGSTON_MENUS
 	VALUE_PAIR *pair_ptr;
+#endif
 
+	log_open(L_AUTH);
+	
+#ifdef USE_LIVINGSTON_MENUS
 	/*
 	 * If the request is processing a menu, service it here.
 	 */
-	if ((pair_ptr = pairfind(authreq->request, DA_STATE)) != NULL &&
+	if ((pair_ptr = avl_find(radreq->request, DA_STATE)) != NULL &&
 	    strncmp(pair_ptr->strvalue, "MENU=", 5) == 0) {
-	    process_menu(authreq, activefd, authreq->digest);
+	    process_menu(radreq, activefd);
 	    return 0;
 	}
 #endif
 
-	m.req = authreq;
+	m.req = radreq;
 	m.activefd = activefd;
 	m.user_check = NULL;
 	m.user_reply = NULL;
@@ -762,8 +745,9 @@ rad_authenticate(authreq, activefd)
 	m.check_pair = NULL;
 	m.timeout_pair = NULL;
 	m.user_msg   = NULL;
+	obstack_init(&m.msg_stack);
 	/*FIXME: this should have been cached by rad_auth_init */
-	m.namepair = pairfind(m.req->request, DA_USER_NAME);
+	m.namepair = avl_find(m.req->request, DA_USER_NAME);
 
 	debug(1, ("auth: %s", m.namepair->strvalue)); 
 	m.state = as_init;
@@ -787,7 +771,7 @@ rad_authenticate(authreq, activefd)
 			default:
 				abort();
 			}
-			if (p = pairfind(p, sp->attr))
+			if (p = avl_find(p, sp->attr))
 				m.check_pair = p;
 			else {
 				m.state = sp->next;
@@ -801,11 +785,12 @@ rad_authenticate(authreq, activefd)
 	}
 
 	/* Cleanup */
-	pairfree(m.user_check);
-	pairfree(m.user_reply);
-	pairfree(m.proxy_pairs);
+	avl_free(m.user_check);
+	avl_free(m.user_reply);
+	avl_free(m.proxy_pairs);
 	if (m.user_msg)
-		free_string(m.user_msg);
+		free(m.user_msg);
+	obstack_free(&m.msg_stack, NULL);
 	bzero(m.userpass, sizeof(m.userpass));
 	return 0;
 }
@@ -818,19 +803,20 @@ rad_authenticate(authreq, activefd)
 #else
 # define newstate(s) m->state = s
 #endif
-
+				
+	
 void
 sfn_init(m)
 	MACH *m;
 {
 	int proxied = 0;
-	AUTH_REQ *authreq = m->req;
+	RADIUS_REQ *radreq = m->req;
 	VALUE_PAIR *pair_ptr;
 	
 	/*
 	 *	Move the proxy_state A/V pairs somewhere else.
 	 */
-	pairmove2(&m->proxy_pairs, &authreq->request, DA_PROXY_STATE);
+	avl_move_attr(&m->proxy_pairs, &radreq->request, DA_PROXY_STATE);
 
 	/*
 	 * If this request got proxied to another server, we need
@@ -838,22 +824,22 @@ sfn_init(m)
 	 * Auth-Reject for fail. We also need to add the reply
 	 * pairs from the server to the initial reply.
 	 */
-	if (authreq->server_code == PW_AUTHENTICATION_REJECT ||
-	    authreq->server_code == PW_AUTHENTICATION_ACK) {
-		m->user_check = create_pair(DA_AUTH_TYPE, 0, NULL, 0);
+	if (radreq->server_code == RT_AUTHENTICATION_REJECT ||
+	    radreq->server_code == RT_AUTHENTICATION_ACK) {
+		m->user_check = avp_create(DA_AUTH_TYPE, 0, NULL, 0);
 		proxied = 1;
 	}
-	if (authreq->server_code == PW_AUTHENTICATION_REJECT)
+	if (radreq->server_code == RT_AUTHENTICATION_REJECT)
 		m->user_check->lvalue = DV_AUTH_TYPE_REJECT;
-	if (authreq->server_code == PW_AUTHENTICATION_ACK)
+	if (radreq->server_code == RT_AUTHENTICATION_ACK)
 		m->user_check->lvalue = DV_AUTH_TYPE_ACCEPT;
 
-	if (authreq->server_reply) {
-		m->user_reply = authreq->server_reply;
-		authreq->server_reply = NULL;
+	if (radreq->server_reply) {
+		m->user_reply = radreq->server_reply;
+		radreq->server_reply = NULL;
 	}
 
-	if (pair_ptr = pairfind(authreq->request, DA_CALLING_STATION_ID)) 
+	if (pair_ptr = avl_find(radreq->request, DA_CALLING_STATION_ID)) 
 		m->clid = pair_ptr->strvalue;
 	else
 		m->clid = _("unknown");
@@ -861,60 +847,133 @@ sfn_init(m)
 	/*
 	 * Get the user from the database
 	 */
-	if (!proxied &&
-	    user_find(m->namepair->strvalue, authreq->request,
-		      &m->user_check, &m->user_reply) != 0) {
+	if (user_find(m->namepair->strvalue, radreq,
+		      &m->user_check, &m->user_reply) != 0
+	    && !proxied) {
 
-		auth_log(m, _("Invalid user"), NULL, NULL, NULL);
+		if (is_log_mode(m, RLOG_AUTH)) 
+			auth_log(m, _("Invalid user"), NULL, NULL, NULL);
 
 		/* Send reject packet with proxy-pairs as a reply */
 		newstate(as_reject);
-		pairfree(m->user_reply);
+		avl_free(m->user_reply);
 		m->user_reply = m->proxy_pairs;
 		m->proxy_pairs = NULL;
 	}
 }
 
+void
+sfn_eval_reply(m)
+	MACH *m;
+{
+	VALUE_PAIR *p;
+	int errcnt = 0;
+	
+	for (p = m->user_reply; p; p = p->next) {
+		if (p->eval) {
+			Datatype type;
+			Datum datum;
+			
+			if (interpret(p->strvalue, m->req, &type, &datum)) {
+				errcnt++;
+				continue;
+			}
+			switch (type) {
+			case Integer:
+				p->lvalue = datum.ival;
+				break;
+			case String:
+				p->strvalue = make_string(datum.sval);
+				p->strlength = strlen(p->strvalue);
+				break;
+			default:
+				abort();
+			}
+			p->eval = 0;
+		}
+	}
+	if (errcnt)
+		newstate(as_reject);
+}		
+
+void
+sfn_scheme(m)
+	MACH *m;
+{
+#ifdef USE_SERVER_GUILE
+	VALUE_PAIR *p;
+	VALUE_PAIR *reply;
+	
+	if (!use_guile) {
+		radlog(L_ERR,
+		       _("Guile authentication disabled in config"));
+		newstate(as_reject);
+		return;
+	}
+
+	reply = avl_dup(m->user_reply);
+	for (p = avl_find(reply, DA_SCHEME_PROCEDURE);
+	     p;
+	     p = avl_find(p->next, DA_SCHEME_PROCEDURE)) {
+		if (scheme_auth(p->strvalue,
+				m->req, m->user_check, &m->user_reply)) {
+			newstate(as_reject);
+			break;
+		}
+	}
+	avl_delete(&m->user_reply, DA_SCHEME_PROCEDURE);
+	avl_free(reply);
+#else
+	radlog(L_ERR,
+	       _("Guile authentication not available"));
+	newstate(as_reject);
+	return;
+#endif
+}
+
 void
 sfn_validate(m)
 	MACH *m;
 {
-	AUTH_REQ *authreq = m->req;
-	VALUE_PAIR *p;
+	RADIUS_REQ *radreq = m->req;
 	int rc;
 	
 	/*
 	 * Validate the user
 	 */
-	if ((rc = check_expiration(m->user_check, &m->user_msg)) >= 0) {
-		rc = rad_check_password(authreq, m->activefd,
-					m->user_check,
-					m->namepair, authreq->digest,
-					&m->user_msg, m->userpass);
+	if ((rc = check_expiration(m)) >= 0) {
+		rc = rad_check_password(radreq,
+					m->user_check, m->namepair,
+					&m->user_msg,
+					m->userpass);
 
 		if (rc != AUTH_OK) { 
-			stat_inc(auth, authreq->ipaddr, num_rejects);
+			stat_inc(auth, radreq->ipaddr, num_rejects);
 			newstate(as_reject);
-			
-			switch (rc) {
-			case AUTH_REJECT:
-				auth_log(m, _("Rejected"), NULL, NULL, NULL);  
-				return;
+			auth_format_msg(m, MSG_ACCESS_DENIED);
+
+			if (is_log_mode(m, RLOG_AUTH)) {
+				switch (rc) {
+				case AUTH_REJECT:
+					auth_log(m, _("Rejected"),
+						 NULL, NULL, NULL);  
+					return;
 				
-			case AUTH_NOUSER:
-				auth_log(m, _("Invalid user"), NULL,
-					 NULL, NULL);
-				return;
+				case AUTH_NOUSER:
+					auth_log(m, _("Invalid user"),
+						 NULL, NULL, NULL);
+					return;
 				
-			case AUTH_FAIL:
-				break;
+				case AUTH_FAIL:
+					break;
+				}
 			}
 		}		
-              /*if (p = pairfind(m->user_reply, DA_REPLY_MESSAGE))
+              /*if (p = avl_find(m->user_reply, DA_REPLY_MESSAGE))
 			m->user_msg = dup_string(p->strvalue);*/
 	}
 
-	pairmove2(&m->user_reply, &m->proxy_pairs, DA_PROXY_STATE);
+	avl_move_attr(&m->user_reply, &m->proxy_pairs, DA_PROXY_STATE);
 
 	if (rc != AUTH_OK) {
 		/*
@@ -940,11 +999,6 @@ sfn_service(m)
 	 */
 	if (m->check_pair->lvalue != DV_SERVICE_TYPE_AUTHENTICATE_ONLY)
 		return;
-	if (is_log_mode(m, RLOG_AUTH)) {
-		auth_log(m, _("Authentication OK"),
-			 is_log_mode(m, RLOG_AUTH_PASS) ? m->userpass : NULL,
-			 NULL, NULL);
-	}
 	newstate(as_ack);
 }
 
@@ -952,7 +1006,8 @@ void
 sfn_disable(m)
 	MACH *m;
 {
-	if (check_disable(m->namepair->strvalue, &m->user_msg)) {
+	if (get_deny(m->namepair->strvalue)) {
+		auth_format_msg(m, MSG_ACCOUNT_CLOSED);
 		auth_log(m, _("Account disabled"), NULL, NULL, NULL);
 		newstate(as_reject);
 	}
@@ -965,7 +1020,7 @@ sfn_service_type(m)
 	if (m->check_pair->lvalue == DV_SERVICE_TYPE_AUTHENTICATE_ONLY) {
 		auth_log(m, _("Login rejected"), NULL,
 			 _("Authenticate only user"), NULL);
-		m->user_msg = make_string(_("\r\nAccess denied\r\n"));
+		auth_format_msg(m, MSG_ACCESS_DENIED);
 		newstate(as_reject);
 	}
 }
@@ -979,8 +1034,7 @@ sfn_realmuse(m)
 	
 	if (rad_check_realm(m->req->realm) == 0)
 		return;
-	m->user_msg = make_string(
-		_("\r\nRealm quota exceeded - access denied\r\n"));
+	auth_format_msg(m, MSG_REALM_QUOTA);
 	auth_log(m, _("Login failed"), NULL,
 		 _("realm quota exceeded for "), m->req->realm);
 	newstate(as_reject);
@@ -991,29 +1045,27 @@ sfn_simuse(m)
 	MACH *m;
 {
 	char  name[AUTH_STRING_LEN];
-	int   rc;
-	char  umsg[AUTH_STRING_LEN];
-
+	int rc;
+	int count;
+	
 	strip_username(strip_names,
 		       m->namepair->strvalue, m->user_check, name);
-	if ((rc = rad_check_multi(name, m->req->request,
-				  m->check_pair->lvalue)) == 0)
+	rc = rad_check_multi(name, m->req->request,
+			     m->check_pair->lvalue, &count);
+	avl_add_pair(&m->user_reply,
+		     avp_create(DA_SIMULTANEOUS_USE, 0, NULL, count));
+	if (!rc)
 		return;
-	
-	if (m->check_pair->lvalue > 1) {
-		radsprintf(umsg, sizeof(umsg),
-	      _("\r\nYou are already logged in %d times  - access denied\r\n"),
-			(int)m->check_pair->lvalue);
-		m->user_msg = make_string(umsg);
-	} else {
-		m->user_msg = make_string(
-		      _("\r\nYou are already logged in - access denied\r\n"));
-	}
+
+	auth_format_msg(m,
+			(m->check_pair->lvalue > 1) ?
+			MSG_MULTIPLE_LOGIN : MSG_SECOND_LOGIN);
 
 	radlog(L_WARN,
-	       _("Multiple logins: [%s] (from nas %s) max. %ld%s"),
+	       _("Multiple logins: [%s] CLID %s (from nas %s) max. %ld%s"),
 	       m->namepair->strvalue,
-	       nas_name2(m->req),
+               m->clid,		
+	       nas_request_to_name(m->req),
 	       m->check_pair->lvalue,
 	       rc == 2 ? _(" [MPP attempt]") : "");
 	newstate(as_reject);
@@ -1024,10 +1076,10 @@ timeout_pair(m)
 	MACH *m;
 {
 	if (!m->timeout_pair &&
-	    !(m->timeout_pair = pairfind(m->user_reply, DA_SESSION_TIMEOUT))) {
-		m->timeout_pair = create_pair(DA_SESSION_TIMEOUT,
+	    !(m->timeout_pair = avl_find(m->user_reply, DA_SESSION_TIMEOUT))) {
+		m->timeout_pair = avp_create(DA_SESSION_TIMEOUT,
 					      0, NULL, 0);
-		pairadd(&m->user_reply, m->timeout_pair);
+		avl_add_pair(&m->user_reply, m->timeout_pair);
 	}
 	return m->timeout_pair;
 }
@@ -1046,12 +1098,11 @@ sfn_time(m)
 		/*
 		 * User called outside allowed time interval.
 		 */
-		m->user_msg = make_string(
-		      _("You are calling outside your allowed timespan\r\n"));
+		auth_format_msg(m, MSG_TIMESPAN_VIOLATION);
 		radlog(L_ERR,
        _("Outside allowed timespan: [%s] (from nas %s) time allowed: %s"),
 		       m->namepair->strvalue,
-		       nas_name2(m->req),
+		       nas_request_to_name(m->req),
 		       m->check_pair->strvalue);
 		newstate(as_reject);
 	} else if (rc == 0) {
@@ -1066,29 +1117,6 @@ sfn_time(m)
 	}
 }
 
-/*ARGSUSED*/
-void
-sfn_ttl(m)
-	MACH *m;
-{
-#ifdef USE_NOTIFY	
-	long r;
-	if (timetolive(m->namepair->strvalue, &r) == 0) {
-		if (r > 0) {
-			timeout_pair(m)->lvalue = r;
-		} else {
-			auth_log(m,
-				 _("Zero time to live"),
-				 NULL, NULL, NULL);
-
-			m->user_msg = make_string(
-			 _("\r\nSorry, your account is currently closed\r\n"));
-			newstate(as_reject);
-		}
-	}
-#endif
-}
-
 void
 sfn_ipaddr(m)
 	MACH *m;
@@ -1096,34 +1124,34 @@ sfn_ipaddr(m)
 	VALUE_PAIR *p, *tmp, *pp;
 	
 	/* Assign an IP if necessary */
-	if (!pairfind(m->user_reply, DA_FRAMED_IP_ADDRESS)) {
+	if (!avl_find(m->user_reply, DA_FRAMED_IP_ADDRESS)) {
 #if 0
 		/* **************************************************
 		 * Keep it here until IP allocation is ready
 		 */
 		if (p = alloc_ip_pair(m->namepair->strvalue, m->req))
-			pairadd(&m->user_reply, p);
+			avl_add_pair(&m->user_reply, p);
 		else
 #endif	
-		if (p = pairfind(m->req->request, DA_FRAMED_IP_ADDRESS)) {
+		if (p = avl_find(m->req->request, DA_FRAMED_IP_ADDRESS)) {
 			/* termserver hint */
-			pairadd(&m->user_reply, pairdup(p));
-			if (p = pairfind(m->req->request,
+			avl_add_pair(&m->user_reply, avp_dup(p));
+			if (p = avl_find(m->req->request,
 					 DA_ADD_PORT_TO_IP_ADDRESS))
-				pairadd(&m->user_reply, pairdup(p));
+				avl_add_pair(&m->user_reply, avp_dup(p));
 		}
 	}
 	
-	if ((p = pairfind(m->user_reply, DA_FRAMED_IP_ADDRESS)) &&
-	    (tmp = pairfind(m->user_reply, DA_ADD_PORT_TO_IP_ADDRESS)) &&
+	if ((p = avl_find(m->user_reply, DA_FRAMED_IP_ADDRESS)) &&
+	    (tmp = avl_find(m->user_reply, DA_ADD_PORT_TO_IP_ADDRESS)) &&
 	    tmp->lvalue &&
-	    (pp = pairfind(m->req->request, DA_NAS_PORT_ID)))
+	    (pp = avl_find(m->req->request, DA_NAS_PORT_ID)))
 		/* NOTE: This only works because IP numbers are stored in
 		 * host order throughout the program.
 		 */
 		p->lvalue += pp->lvalue;
 		
-	pairdelete(&m->user_reply, DA_ADD_PORT_TO_IP_ADDRESS);
+	avl_delete(&m->user_reply, DA_ADD_PORT_TO_IP_ADDRESS);
 }
 
 void
@@ -1131,7 +1159,7 @@ sfn_exec_wait(m)
 	MACH *m;
 {
 	if (radius_exec_program(m->check_pair->strvalue,
-				m->req->request,
+				m->req,
 				&m->user_reply,
 				1,
 				&m->user_msg) != 0) {
@@ -1143,9 +1171,7 @@ sfn_exec_wait(m)
 
 		newstate(as_reject);
 
-		if (!m->user_msg)
-			m->user_msg = make_string(
-			     _("\r\nAccess denied (external check failed)."));
+		auth_format_msg(m, MSG_ACCESS_DENIED);
 
 		if (is_log_mode(m, RLOG_AUTH)) {
 			auth_log(m, _("Login incorrect"),
@@ -1160,7 +1186,7 @@ sfn_exec_nowait(m)
 {
 	/*FIXME: do we need to pass user_reply here? */
 	radius_exec_program(m->check_pair->strvalue,
-			    m->req->request, &m->user_reply,
+			    m->req, &m->user_reply,
 			    0, NULL);
 }
 
@@ -1183,7 +1209,7 @@ sfn_cleanup_cbkid(m)
 	int *ip;
 
 	for (ip = delete_pairs; *ip; ip++)
-		pairdelete(&m->user_reply, *ip);
+		avl_delete(&m->user_reply, *ip);
 }
 
 void
@@ -1195,7 +1221,7 @@ sfn_menu(m)
 	char state_value[MAX_STATE_VALUE];
 		
 	msg = get_menu(m->check_pair->strvalue);
-	radsprintf(state_value, sizeof(state_value),
+	snprintf(state_value, sizeof(state_value),
 		   "MENU=%s", m->check_pair->strvalue);
 	send_challenge(m->req, msg, state_value, m->activefd);
 	
@@ -1214,10 +1240,10 @@ sfn_ack(m)
 	
 	stat_inc(auth, m->req->ipaddr, num_accepts);
 
-	rad_send_reply(PW_AUTHENTICATION_ACK,
+	rad_send_reply(RT_AUTHENTICATION_ACK,
 		       m->req,
 		       m->user_reply,
-		       m->user_msg,
+		       auth_finish_msg(m),
 		       m->activefd);
 	
 	if (is_log_mode(m, RLOG_AUTH)) {
@@ -1233,18 +1259,41 @@ sfn_ack(m)
 	}
 }
 
-
 void
 sfn_reject(m)
 	MACH *m;
 {
 	debug(1, ("REJECT: %s", m->namepair->strvalue));
-	rad_send_reply(PW_AUTHENTICATION_REJECT,
+	rad_send_reply(RT_AUTHENTICATION_REJECT,
 		       m->req,
 		       m->user_reply,
-		       m->user_msg,
+		       auth_finish_msg(m),
 		       m->activefd);
 	stat_inc(auth, m->req->ipaddr, num_rejects);
 }
 
-
+void
+req_decrypt_password(password, req, pair)
+	char *password;
+	RADIUS_REQ *req;
+	VALUE_PAIR *pair; /* Password pair */
+{
+	NAS *nas;
+	char *s;
+	
+	if (!pair) {
+		pair = avl_find(req->request, DA_PASSWORD);
+		if (!pair)
+			return;
+	}
+	/* Determine whether we need to use broken decoding */
+	nas = nas_request_to_nas(req);
+	if (nas
+            && (s = envar_lookup(nas->args, "broken_pass")) != NULL
+	    && s[0] == '1')
+		decrypt_password_broken(password, pair,
+					req->vector, req->secret);
+	else
+		decrypt_password(password, pair,
+				 req->vector, req->secret);
+}
