@@ -1,25 +1,24 @@
 /* This file is part of GNU RADIUS.
- * Copyright (C) 2000,2001, Sergey Poznyakoff
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
- */
-#define RADIUS_MODULE 7
+   Copyright (C) 2000,2001, Sergey Poznyakoff
+  
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+  
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+  
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation,
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+
+#define RADIUS_MODULE_EXEC_C
 #ifndef lint
 static char rcsid[] =
-"@(#) $Id: exec.c,v 1.10 2001/03/15 10:54:47 gray Exp $"; 
+"@(#) $Id: exec.c,v 1.18 2001/11/27 12:13:36 gray Exp $"; 
 #endif
 
 #ifdef HAVE_CONFIG_H
@@ -44,10 +43,7 @@ static char rcsid[] =
 #include <string.h>
 #include <syslog.h>
 #include <radiusd.h>
-
-
-#define MAXARGS 64
-
+#include <obstack1.h>
 
 /*
  *	Execute a program on successful authentication.
@@ -56,9 +52,9 @@ static char rcsid[] =
  *
  */
 int
-radius_exec_program(cmd, request, reply, exec_wait, user_msg)
+radius_exec_program(cmd, req, reply, exec_wait, user_msg)
 	char *cmd;
-	VALUE_PAIR *request;
+	RADIUS_REQ *req;
 	VALUE_PAIR **reply;
 	int exec_wait;
 	char **user_msg;
@@ -66,8 +62,6 @@ radius_exec_program(cmd, request, reply, exec_wait, user_msg)
 	int p[2];
 	RETSIGTYPE (*oldsig)();
 	pid_t pid;
-	int argc;
-	char *argv[MAXARGS];
 	int n;
 	char *ptr, *errp;
 	int status;
@@ -76,7 +70,6 @@ radius_exec_program(cmd, request, reply, exec_wait, user_msg)
 	int line_num;
 	char buffer[RAD_BUFFER_SIZE];
 	struct passwd *pwd;
-	struct group  *grp;
 #if 0
 	int saved_uid, saved_gid;
 #endif
@@ -99,28 +92,6 @@ radius_exec_program(cmd, request, reply, exec_wait, user_msg)
 		       config.exec_user);
 		return -1;
 	}
-	grp = getgrnam(config.exec_group);
-	if (!grp) {
-		radlog(L_ERR,
-    _("radius_exec_program(): won't execute, no such group: %s"),
-		       config.exec_group);
-		return -1;
-	} else if (pwd->pw_gid != grp->gr_gid) {
-		int ok = 0;
-		char **mp;
-
-		for (mp = grp->gr_mem; *mp; mp++)
-			if (strcmp(*mp, pwd->pw_name) == 0) {
-				ok++;
-				break;
-			}
-		if (!ok) {
-			radlog(L_ERR,
-     _("radius_exec_program(): won't execute, user %s not in group %s"),
-			       pwd->pw_name, grp->gr_name);
-			return -1;
-		}
-	}
 	
 	if (exec_wait) {
 		if (pipe(p) != 0) {
@@ -128,6 +99,7 @@ radius_exec_program(cmd, request, reply, exec_wait, user_msg)
 			p[0] = p[1] = 0;
 			return -1;
 		}
+
 		if ((oldsig = signal(SIGCHLD, SIG_DFL)) == SIG_ERR) {
 			radlog(L_ERR|L_PERROR, _("can't reset SIGCHLD"));
 			return -1;
@@ -135,24 +107,20 @@ radius_exec_program(cmd, request, reply, exec_wait, user_msg)
 	}
 
 	if ((pid = fork()) == 0) {
+		int argc;
+		char **argv;
+		struct obstack s;
+
+		obstack_init(&s);
+		
 		/* child branch */
-		ptr = radius_xlate(buffer, sizeof(buffer), cmd, request, *reply);
+		ptr = radius_xlate(&s, cmd, req, reply ? *reply : NULL);
 
 		debug(1,
 			("command line: %s", ptr));
 
-		argc = 0;
-		for (ptr = strtok(ptr, " \t"); ptr; ptr=strtok(NULL, " \t")) {
-			if (argc > MAXARGS) {
-				radlog(L_ERR,
-				    _("radius_exec_program(): too many arguments"));
-				return -1;
-			}
-			debug(10, ("argv[%d] = %s", argc, ptr));
-			argv[argc++] = ptr;
-		}
-		argv[argc++] = NULL;
-
+		argcv_get(ptr, "", &argc, &argv);
+		
 		if (exec_wait) {
 			if (close(p[0]))
 				radlog(L_ERR|L_PERROR, _("can't close pipe"));
@@ -170,7 +138,7 @@ radius_exec_program(cmd, request, reply, exec_wait, user_msg)
 		saved_gid = getegid();
 		#endif
 		
-		if (grp->gr_gid && setgid(grp->gr_gid)) {
+		if (pwd->pw_gid != 0 && setgid(pwd->pw_gid)) {
 			radlog(L_ERR|L_PERROR,
 			       _("setgid(%d) failed"), pwd->pw_gid);
 		}
@@ -225,7 +193,7 @@ radius_exec_program(cmd, request, reply, exec_wait, user_msg)
 			radlog(L_ERR,
 			    _("<stdout of %s>:%d: %s"),
 			    cmd, line_num, errp);
-			pairfree(vp);
+			avl_free(vp);
 			vp = NULL;
 		}
 	}
@@ -234,7 +202,7 @@ radius_exec_program(cmd, request, reply, exec_wait, user_msg)
 	/*close(p[0]);*/
 
 	if (vp) {
-		pairmove(reply, &vp);
+		avl_merge(reply, &vp);
 	}
 
 	
