@@ -1,26 +1,24 @@
 /* This file is part of GNU RADIUS.
- * Copyright (C) 2000, Sergey Poznyakoff
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
- */
+   Copyright (C) 2000, Sergey Poznyakoff
+  
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+  
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+  
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation,
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 
-#define RADIUS_MODULE 12
+#define RADIUS_MODULE_PROXY_C
 #ifndef lint
 static char rcsid[] =
-"@(#) $Id: proxy.c,v 1.10 2001/03/31 14:20:32 gray Exp $";
+"@(#) $Id: proxy.c,v 1.23 2001/11/06 12:39:20 gray Exp $";
 #endif
 
 #ifdef HAVE_CONFIG_H
@@ -51,25 +49,13 @@ static u_char *send_buffer = (u_char *)i_send_buffer;
 
 static PROXY_ID *proxy_id;
 
-static int allowed[] = {
-	DA_SERVICE_TYPE,
-	DA_FRAMED_PROTOCOL,
-	DA_FILTER_ID,
-	DA_FRAMED_MTU,
-	DA_FRAMED_COMPRESSION,
-	DA_LOGIN_SERVICE,
-	DA_REPLY_MESSAGE,
-	DA_SESSION_TIMEOUT,
-	DA_IDLE_TIMEOUT,
-	DA_PORT_LIMIT,
-	0,
-};
-
 static void random_vector(char *vector);
-static VALUE_PAIR *proxy_addinfo(AUTH_REQ *authreq, int proxy_id, UINT4 remip);
-void proxy_addrequest(AUTH_REQ *authreq);
-static void passwd_recode(char *secret_key, char *vector,
-			  char *pw_digest, VALUE_PAIR *pair);
+static VALUE_PAIR *proxy_addinfo(RADIUS_REQ *radreq, int proxy_id, UINT4 remip);
+void proxy_addrequest(RADIUS_REQ *radreq);
+static void passwd_recode(VALUE_PAIR *pair,
+			  char *new_secret, char *new_vector, RADIUS_REQ *req);
+static void rad_send_request(int fd, UINT4 ipaddr, int port, int id,
+			     char *secret, RADIUS_REQ *req);
 
 
 /* ************************************************************************* */
@@ -119,7 +105,7 @@ proxy_cleanup()
 	prev = NULL;
 	for (p = proxy_id; p; ) {
 		next = p->next;
-		if (!client_find(p->ipaddr)) {
+		if (!client_lookup_ip(p->ipaddr)) {
 			if (prev) 
 				prev->next = next;
 			else
@@ -139,8 +125,8 @@ proxy_cleanup()
  */
 /*ARGSUSED*/
 int
-rad_proxy(authreq, activefd)
-	AUTH_REQ *authreq;
+rad_proxy(radreq, activefd)
+	RADIUS_REQ *radreq;
 	int activefd;
 {
 	void *ptr;
@@ -148,11 +134,11 @@ rad_proxy(authreq, activefd)
 	/*
 	 *	Copy the static data into malloc()ed memory.
 	 */
-	ptr = emalloc(authreq->data_len);
+	ptr = emalloc(radreq->data_len);
 	debug(1,("allocated ptr %p", ptr));
-	memcpy(ptr, authreq->data, authreq->data_len);
-	authreq->data = ptr;
-	authreq->data_alloced = 1;
+	memcpy(ptr, radreq->data, radreq->data_len);
+	radreq->data = ptr;
+	radreq->data_alloced = 1;
 
 	return 0;
 }
@@ -161,9 +147,7 @@ rad_proxy(authreq, activefd)
 /* ************************************************************************* */
 /* Reply functions. Possibly these should go to libclient? */
 
-/*
- *	Generate a random vector.
- */
+/* Generate a random vector. */
 static void
 random_vector(vector)
 	char *vector;
@@ -181,15 +165,13 @@ random_vector(vector)
 }
 
 void
-rad_send_request(fd, ipaddr, port, id, code, pw_digest, secret_key, request)
+rad_send_request(fd, ipaddr, port, id, secret, req)
 	int   fd;
 	UINT4 ipaddr;
 	int   port;
 	int   id;
-	int   code;
-	char  *pw_digest;
-	char  *secret_key;
-	VALUE_PAIR *request;
+	char  *secret;
+	RADIUS_REQ *req;
 {
 	AUTH_HDR		*auth;
 	VALUE_PAIR		*vp;
@@ -208,9 +190,9 @@ rad_send_request(fd, ipaddr, port, id, code, pw_digest, secret_key, request)
 	random_vector(vector);
 	auth = (AUTH_HDR *)send_buffer;
 	memset(auth, 0, sizeof(AUTH_HDR));
-	auth->code = code;
+	auth->code = req->code;
 	auth->id = id;
-	if (auth->code == PW_AUTHENTICATION_REQUEST)
+	if (auth->code == RT_AUTHENTICATION_REQUEST)
 		memcpy(auth->vector, vector, AUTH_VECTOR_LEN);
 
 	total_length = AUTH_HDR_LEN;
@@ -219,14 +201,11 @@ rad_send_request(fd, ipaddr, port, id, code, pw_digest, secret_key, request)
 	 *	Put all the attributes into a buffer.
 	 */
 	ptr = auth->data;
-	for (vp = request; vp; vp = vp->next) {
+	for (vp = req->request; vp; vp = vp->next) {
 
-		if (debug_on(10))
-			debug_pair("proxy_send", vp);
+		debug(10, ("proxy_send: %s", format_pair(vp)));
 
-		/*
-		 *	This could be a vendor-specific attribute.
-		 */
+		/* This could be a vendor-specific attribute. */
 		length_ptr = NULL;
 		if ((vendorcode = VENDOR(vp->attribute)) > 0 &&
 		    (vendorpec  = vendor_id_to_pec(vendorcode)) > 0) {
@@ -239,9 +218,6 @@ rad_send_request(fd, ipaddr, port, id, code, pw_digest, secret_key, request)
 			ptr += 4;
 			total_length += 6;
 		} else if (vp->attribute > 0xff) {
-			/*
-			 *	Ignore attributes > 0xff
-			 */
 			continue;
 		} else
 			vendorpec = 0;
@@ -249,13 +225,13 @@ rad_send_request(fd, ipaddr, port, id, code, pw_digest, secret_key, request)
 		*ptr++ = (vp->attribute & 0xFF);
 
 		switch (vp->type) {
-		case PW_TYPE_STRING:
+		case TYPE_STRING:
 			/*
 			 *	Re-encode passwd on the fly.
 			 */
 			if (vp->attribute == DA_PASSWORD)
-				passwd_recode(secret_key, vector,
-					      pw_digest, vp);
+				passwd_recode(vp, secret, vector, req);
+			
                         checkovf(vp->strlength + 2);
 
 			*ptr++ = vp->strlength + 2;
@@ -267,9 +243,9 @@ rad_send_request(fd, ipaddr, port, id, code, pw_digest, secret_key, request)
 
 			ptr += vp->strlength;
 			break;
-		case PW_TYPE_INTEGER:
-		case PW_TYPE_DATE:
-		case PW_TYPE_IPADDR:
+		case TYPE_INTEGER:
+		case TYPE_DATE:
+		case TYPE_IPADDR:
                         checkovf(sizeof(UINT4) + 2); 
 
 			*ptr++ = sizeof(UINT4) + 2;
@@ -286,15 +262,12 @@ rad_send_request(fd, ipaddr, port, id, code, pw_digest, secret_key, request)
 	}
 	auth->length = htons(total_length);
 
-	/*
-	 *	If this is not an authentication request, we
-	 *	need to calculate the md5 hash over the entire packet
-	 *	and put it in the vector.
-	 */
-	if (auth->code != PW_AUTHENTICATION_REQUEST) {
-		len = strlen(secret_key);
+	/* If this is not an authentication request, we	need to calculate
+	   the md5 hash over the entire packet and put it in the vector. */
+	if (auth->code != RT_AUTHENTICATION_REQUEST) {
+		len = strlen(secret);
 		if (total_length + len < sizeof(i_send_buffer)) {
-			strcpy(send_buffer + total_length, secret_key);
+			strcpy(send_buffer + total_length, secret);
 			md5_calc(auth->vector, send_buffer, total_length+len);
 		}
 	}
@@ -325,28 +298,28 @@ ovf:
  *	Add a proxy-pair to the end of the request.
  */
 static VALUE_PAIR *
-proxy_addinfo(authreq, proxy_id, remip)
-	AUTH_REQ *authreq;
+proxy_addinfo(radreq, proxy_id, remip)
+	RADIUS_REQ *radreq;
 	int proxy_id;
 	UINT4 remip;
 {
 	VALUE_PAIR		*proxy_pair, *vp;
 	PROXY_STATE		*proxy_state;
 
-	proxy_pair = alloc_pair();
+	proxy_pair = avp_alloc();
 	proxy_pair->name = "Proxy-State";
 	proxy_pair->attribute = DA_PROXY_STATE;
-	proxy_pair->type = PW_TYPE_STRING;
+	proxy_pair->type = TYPE_STRING;
 	proxy_pair->strlength = sizeof(PROXY_STATE);
 	proxy_pair->strvalue = alloc_string(proxy_pair->strlength);
 	
 	proxy_state = (PROXY_STATE *)proxy_pair->strvalue;
 	proxy_state->ipaddr = myip;
-	proxy_state->id = authreq->id;
+	proxy_state->id = radreq->id;
 	proxy_state->proxy_id = proxy_id;
 	proxy_state->rem_ipaddr = remip;
 
-	for (vp = authreq->request; vp && vp->next; vp = vp->next)
+	for (vp = radreq->request; vp && vp->next; vp = vp->next)
 		;
 	vp->next = proxy_pair;
 	return vp;
@@ -356,48 +329,18 @@ proxy_addinfo(authreq, proxy_id, remip)
  *	Decode a password and encode it again.
  */
 static void
-passwd_recode(secret_key, vector, pw_digest, pass_pair)
-	char *secret_key;
-	char *vector;
-	char *pw_digest;
+passwd_recode(pass_pair, new_secret, new_vector, req)
 	VALUE_PAIR *pass_pair;
+	char *new_secret;
+	char *new_vector;
+	RADIUS_REQ *req;
 {
-	char	passtmp[AUTH_PASS_LEN];
-	char	passwd[AUTH_PASS_LEN];
-	char	md5buf[256];
-	int	i;
-	int	len;
-
-	/*
-	 * Decode. First fixup the password padding it with
-	 * zeroes to next 16-character boundary
-	 */
-	memset(passwd, 0, AUTH_PASS_LEN);
-	memcpy(passwd, pass_pair->strvalue, AUTH_PASS_LEN);
-
-	for(i = 0 ; i < AUTH_PASS_LEN; i++)
-		passwd[i] ^= pw_digest[i];
-
-	/*
-	 *	Encode with new secret.
-	 */
-	memset(passtmp, 0, sizeof(passtmp));
-	len = strlen(secret_key);
-	strcpy(md5buf, secret_key);
-	memcpy(md5buf + len, vector, AUTH_VECTOR_LEN);
-	md5_calc(passtmp, md5buf, len + AUTH_VECTOR_LEN);
-	for (i = 0; i < AUTH_PASS_LEN; i++)
-		passtmp[i] ^= passwd[i];
-
-	/*
-	 *	Copy newly encoded password back to
-	 *	where we got it from.
-	 */
-	if (pass_pair->strlength < AUTH_PASS_LEN) {
-		free_string(pass_pair->strvalue);
-		pass_pair->strvalue = alloc_string(AUTH_PASS_LEN);
-	}
-	memcpy(pass_pair->strvalue, passtmp, AUTH_PASS_LEN);
+	char	password[AUTH_STRING_LEN+1];
+	req_decrypt_password(password, req, pass_pair);
+	free_string(pass_pair->strvalue);
+	encrypt_password(pass_pair, password, new_vector, new_secret);
+	/* Don't let the cleantext hang around */
+	memset(password, 0, AUTH_STRING_LEN);
 }
 
 /* ************************************************************************* */
@@ -411,8 +354,8 @@ passwd_recode(secret_key, vector, pw_digest, pass_pair)
  *
  */
 int
-proxy_send(authreq, activefd)
-	AUTH_REQ *authreq;
+proxy_send(radreq, activefd)
+	RADIUS_REQ *radreq;
 	int activefd;
 {
 	int                     rc;
@@ -425,67 +368,66 @@ proxy_send(authreq, activefd)
 	CLIENT			*client;
 	short			rport;
 	char                    *what;
-	char			pw_digest[16];
 	char                    *secret_key;
 	u_char                  proxy_id;
 	
 	/*
 	 *	Look up name.
 	 */
-	namepair = pairfind(authreq->request, DA_USER_NAME);
+	namepair = avl_find(radreq->request, DA_USER_NAME);
 	if (namepair == NULL)
 		return 0;
 
 	saved_username = estrdup(namepair->strvalue);
 	username = namepair->strvalue;
 
-	/*
-	 *	Find the realm from the _end_ so that we can
-	 *	cascade realms: user@realm1@realm2.
-	 */
+	/* Find the realm from the _end_ so that we can	cascade realms:
+	   user@realm1@realm2. Two special realms are handled separately:
+	   
+	       LOCAL    -- process request locally.
+	       NOREALM  -- handle an empty realm name.
+
+	   A realm with special name DEFAULT is returned by realm_find()
+	   if no other realm name matches. */
 
 	if ((realmname = strrchr(username, '@')) == NULL) {
+		if ((realm = realm_find("NOREALM")) == NULL) {
+			efree(saved_username);
+			return 0;
+		}
+	} else if ((realm = realm_find(realmname + 1)) == NULL) {
+		/* If the realm is not found, we treat it as usual. */
 		efree(saved_username);
 		return 0;
-	}
-
-	/*
-	 *	Now check if we know this realm!
-	 *	If not found, we treat it as usual.
-	 */
-	if ((realm = realm_find(realmname + 1)) == NULL) {
-		efree(saved_username);
-		return 0;
-	}
-
-	/*
-	 *	The special realm LOCAL ?
-	 */
-	if (strcmp(realm->server, "LOCAL") == 0) {
-		if (realm->striprealm &&
-		    ((realmname = strrchr(namepair->strvalue, '@')) != NULL))
+	} else if (strcmp(realm->server, "LOCAL") == 0) {
+		if (realm->striprealm) {
 			*realmname = 0;
+			namepair->strlength = strlen(namepair->strvalue);
+		}
 		efree(saved_username);
 		return 0;
 	}
 
-	if ((client = client_find(realm->ipaddr)) == NULL) {
-		radlog(L_PROXY,
+	if ((client = client_lookup_ip(realm->ipaddr)) == NULL) {
+		radlog(L_PROXY|L_ERR,
 		       _("cannot find secret for server %s in clients file"),
 		       realm->server);
 		efree(saved_username);
 		return 0;
 	}
 
-	if (realm->striprealm)
-		*realmname = 0;
-	realmname++;
-	authreq->realm = make_string(realmname);
-
+	if (realmname) {
+		if (realm->striprealm)
+			*realmname = 0;
+		realmname++;
+		radreq->realm = make_string(realmname);
+	} else
+		radreq->realm = make_string(realm->realm);
+	
 	replace_string(&namepair->strvalue, username);
 	namepair->strlength = strlen(namepair->strvalue);
 
-	if (authreq->code == PW_AUTHENTICATION_REQUEST)
+	if (radreq->code == RT_AUTHENTICATION_REQUEST)
 		rport = realm->auth_port;
 	else
 		rport = realm->acct_port;
@@ -493,8 +435,8 @@ proxy_send(authreq, activefd)
 	secret_key = client->secret;
 	proxy_id = next_proxy_id(client->ipaddr);
 	
-	authreq->server_ipaddr = realm->ipaddr;
-	authreq->server_id = proxy_id;
+	radreq->server_ipaddr = realm->ipaddr;
+	radreq->server_id = proxy_id;
 
 	/*
 	 *	Is this a valid & signed request ?
@@ -503,14 +445,14 @@ proxy_send(authreq, activefd)
 	 *	FIXME: we have already calculated the
 	 *	digest in rad_auth_init()
 	 */
-	switch (authreq->code) {
-	case PW_AUTHENTICATION_REQUEST:
+	switch (radreq->code) {
+	case RT_AUTHENTICATION_REQUEST:
 		what = _("authentication");
-		rc = calc_digest(pw_digest, authreq) != 0;
+		rc = 0;
 		break;
-	case PW_ACCOUNTING_REQUEST:
+	case RT_ACCOUNTING_REQUEST:
 		what = _("accounting");
-		rc = calc_acctdigest(pw_digest, authreq) < 0;
+		rc = calc_acctdigest(radreq) < 0;
 		break;
 	default:
 		what = _("unknown");
@@ -520,7 +462,7 @@ proxy_send(authreq, activefd)
 	if (rc) {
 		radlog(L_NOTICE,
 		       _("%s request from client %s for user %s - Security Breach"),
-		       what, client_name(authreq->ipaddr),
+		       what, client_lookup_name(radreq->ipaddr),
 		       namepair->strvalue);
 		efree(saved_username);
 		return -1;
@@ -529,52 +471,44 @@ proxy_send(authreq, activefd)
 	/*
 	 *	Add PROXY_STATE attribute.
 	 */
-	pp = proxy_addinfo(authreq, proxy_id, realm->ipaddr);
+	pp = proxy_addinfo(radreq, proxy_id, realm->ipaddr);
 
 	/*
 	 *	If there is no DA_CHAP_CHALLENGE attribute but there
 	 *	is a DA_CHAP_PASSWORD we need to add it since we can't
 	 *	use the request authenticator anymore - we changed it.
 	 */
-	if (pairfind(authreq->request, DA_CHAP_PASSWORD) &&
-	    pairfind(authreq->request, DA_CHAP_CHALLENGE) == NULL) {
-		vp = alloc_pair();
+	if (avl_find(radreq->request, DA_CHAP_PASSWORD) &&
+	    avl_find(radreq->request, DA_CHAP_CHALLENGE) == NULL) {
+		vp = avp_alloc();
 		
 		memset(vp, 0, sizeof(VALUE_PAIR));
 
 		vp->name = "CHAP-Challenge";
 		vp->attribute = DA_CHAP_CHALLENGE;
-		vp->type = PW_TYPE_STRING;
+		vp->type = TYPE_STRING;
 		vp->strlength = AUTH_VECTOR_LEN;
 		vp->strvalue = alloc_string(AUTH_VECTOR_LEN);
-		memcpy(vp->strvalue, authreq->vector, AUTH_VECTOR_LEN);
-		pairadd(&authreq->request, vp);
+		memcpy(vp->strvalue, radreq->vector, AUTH_VECTOR_LEN);
+		avl_add_pair(&radreq->request, vp);
 	}
 
 	debug(1, ("Sending %s request of id %d to %lx (server %s:%d)",
 		 what, proxy_id, realm->ipaddr, realm->server, rport));
 
-	/*
-	 *	Now build a new request and send it to the remote radiusd.
-	 */
-	rad_send_request(activefd, realm->ipaddr, rport,
-			 proxy_id, authreq->code,
-			 pw_digest, secret_key,
-			 authreq->request);
+	/* Now build a new request and send it to the remote radiusd. */
+	rad_send_request(activefd, realm->ipaddr, rport, proxy_id,
+			 secret_key, radreq);
 	
-	/*
-	 *	Remove proxy-state from list.
-	 */
+	/* Remove proxy-state from list. */
 	if (pp->next) {
 		VALUE_PAIR *p = pp->next;
 		pp->next = p->next;
 		p->next = NULL;  /* be sure to delete *only* this pair */
-		pairfree(p);
+		avl_free(p);
 	}
 #if 1	
-	/*
-	 *	And restore username.
-	 */
+	/* And restore username. */
 	replace_string(&namepair->strvalue, saved_username);
 	namepair->strlength = strlen(namepair->strvalue);
 #endif
@@ -585,23 +519,22 @@ proxy_send(authreq, activefd)
 
 /* ************************************************************************* */
 /* Functions for finding the matching request in the list of outstanding ones.
- * There seem to be two cases: i) when the remote server retains the
+ * There appear to be two cases: i) when the remote server retains the
  * Proxy-State A/V pair, which seems to correspond to RFC 2865,
  * and ii) when the remote server drops the Proxy-State pair.
  */
 
 struct proxy_data {
 	PROXY_STATE *state;
-	AUTH_REQ    *authreq;
+	RADIUS_REQ    *radreq;
 };
 
 /* proxy_compare_request(): Find matching request based on the information
- * preserved in the Proxy-State pair.
- */
+   preserved in the Proxy-State pair. */
 int
 proxy_compare_request(data, oldreq)
 	struct proxy_data *data;
-	AUTH_REQ *oldreq;
+	RADIUS_REQ *oldreq;
 {
 	debug(10, ("(old=data) id %d %d, ipaddr %#8x %#8x", 
 		oldreq->id,data->state->id,myip,data->state->ipaddr));
@@ -614,62 +547,63 @@ proxy_compare_request(data, oldreq)
 }
 
 /* proxy_compare_request_no_state(): Find matching outstanding request if the
- * server did not retain the Proxy-State pair.
- * miquels@cistron.nl says:
- *	Some servers drop the proxy pair. So
- *      compare in another way if needed.
- *	FIXME: hmmm, perhaps we don't even need Proxy-State
- *	after all!
- */
+   server did not retain the Proxy-State pair.
+   miquels@cistron.nl says:
+  	Some servers drop the proxy pair. So
+        compare in another way if needed.
+  	FIXME: hmmm, perhaps we don't even need Proxy-State
+  	after all! */
 int
 proxy_compare_request_no_state(data, oldreq)
 	struct proxy_data *data;
-	AUTH_REQ *oldreq;
+	RADIUS_REQ *oldreq;
 {
 	debug(10, ("(old=data) id %d %d, ipaddr %#8x %#8x",
 		oldreq->server_id,
-		data->authreq->id,
+		data->radreq->id,
 		oldreq->server_ipaddr,
-		data->authreq->ipaddr));
+		data->radreq->ipaddr));
 			
-	if (data->authreq->ipaddr == oldreq->server_ipaddr &&
-	    data->authreq->id     == oldreq->server_id)
+	if (data->radreq->ipaddr == oldreq->server_ipaddr &&
+	    data->radreq->id     == oldreq->server_id)
 		return 0;
 
 	return 1;
 }
 
 
-/*
- *	We received a response from a remote radius server.
- *	Find the original request, then return.
- *	Returns:   0 proxy found
- *		  -1 error don't reply
- */
 int
-proxy_receive(authreq, activefd)
-	AUTH_REQ        *authreq;
+select_allowed(unused, pair)
+	void *unused;
+	VALUE_PAIR *pair;
+{
+	return pair->prop & AP_PROPAGATE;
+}
+
+/* Called when a response from a remote radius server has been received.
+   The function finds the original request and replaces all fields in
+   radreq, except `request' with the original data.
+   Return:   0 proxy found
+  	    -1 error don't reply */
+int
+proxy_receive(radreq, activefd)
+	RADIUS_REQ        *radreq;
 	int             activefd;
 {
 	VALUE_PAIR	*vp, *proxy_state_pair, *prev, *x;
 	VALUE_PAIR	*allowed_pairs;
-	AUTH_REQ	*oldreq;
+	RADIUS_REQ	*oldreq;
 	PROXY_STATE	*state;
 	struct proxy_data data;
-	int             i;
 	
-	/*
-	 *	FIXME: calculate md5 checksum!
-	 */
+	/* FIXME: calculate md5 checksum! */
 
-	/*
-	 *	Find the last PROXY_STATE attribute.
-	 */
+	/* Find the last PROXY_STATE attribute. */
 
 	oldreq  = NULL;
 	proxy_state_pair = x = prev = NULL;
 
-	for (vp = authreq->request; vp; vp = vp->next) {
+	for (vp = radreq->request; vp; vp = vp->next) {
 		if (vp->attribute == DA_PROXY_STATE) {
 			prev = x;
 			proxy_state_pair = vp;
@@ -687,18 +621,16 @@ proxy_receive(authreq, activefd)
 		 state->proxy_id,
 		 state->rem_ipaddr));
 
-	/*
-	 *	Now find matching request in the list of outstanding requests.
-	 */
+	/* Find matching request in the list of outstanding requests. */
 	data.state = state;
-	data.authreq = authreq;
+	data.radreq = radreq;
 
-	debug(1, ("Compare: myip %08x, authreq->id %d, authreq->ipaddr %08x",
-		 myip, authreq->id, authreq->ipaddr));
+	debug(1, ("Compare: myip %08x, radreq->id %d, radreq->ipaddr %08x",
+		 myip, radreq->id, radreq->ipaddr));
 	
 	if (state) {
-		if (state->proxy_id   == authreq->id &&
-		    state->rem_ipaddr == authreq->ipaddr) {
+		if (state->proxy_id   == radreq->id &&
+		    state->rem_ipaddr == radreq->ipaddr) {
 			oldreq = scan_request_list(R_PROXY,
 						   proxy_compare_request,
 						   &data);
@@ -712,55 +644,47 @@ proxy_receive(authreq, activefd)
 	}
 	
 	if (oldreq == NULL) {
-		radlog(L_PROXY,
+		radlog(L_PROXY|L_ERR,
 		       _("Unrecognized proxy reply from server %s - ID %d"),
-		       client_name(authreq->ipaddr), authreq->id);
+		       client_lookup_name(radreq->ipaddr), radreq->id);
 		return -1;
 	}
 
-	/*
-	 *	Remove proxy pair from list.
-	 */
+	/* Remove proxy pair from the list. */
 	if (proxy_state_pair) {
 		if (prev)
 			prev->next = proxy_state_pair->next;
 		else
-			authreq->request = proxy_state_pair->next;
+			radreq->request = proxy_state_pair->next;
           	proxy_state_pair->next = NULL;
-	        pairfree(proxy_state_pair);
+	        avl_free(proxy_state_pair);
 	}
 
-	/*
-	 *	Only allow some attributes to be propagated from
-	 *	the remote server back to the NAS, for security.
-	 */
+	/* Only allow some attributes to be propagated from
+	   the remote server back to the NAS, for security. */
 	allowed_pairs = NULL;
-	for(i = 0; allowed[i]; i++)
-		pairmove2(&allowed_pairs, &authreq->request, allowed[i]);
-	pairfree(authreq->request);
+	avl_move_pairs(&allowed_pairs, &radreq->request, select_allowed, NULL);
+	avl_free(radreq->request);
 
-	/*
-	 *	Now rebuild the AUTHREQ struct, so that the
-	 *	normal functions can process it. Take care not to modify
-	 *      oldreq!
-	 */
+	/* Rebuild the RADIUS_REQ struct, so that the normal functions
+	   can process it. Take care not to modify oldreq! */
 	
-	authreq->server_reply = allowed_pairs;
-	authreq->validated    = 1;
-	authreq->server_code  = authreq->code;
-	authreq->code         = oldreq->code;
+	radreq->server_reply = allowed_pairs;
+	radreq->validated    = 1;
+	radreq->server_code  = radreq->code;
+	radreq->code         = oldreq->code;
 
-	authreq->ipaddr       = oldreq->ipaddr;
-	authreq->udp_port     = oldreq->udp_port;
-	authreq->id           = oldreq->id;
-	memcpy(authreq->vector, oldreq->vector, sizeof authreq->vector);
-	memcpy(authreq->secret, oldreq->secret, sizeof authreq->secret);
-	authreq->request      = paircopy(oldreq->request);
+	radreq->ipaddr       = oldreq->ipaddr;
+	radreq->udp_port     = oldreq->udp_port;
+	radreq->id           = oldreq->id;
+	memcpy(radreq->vector, oldreq->vector, sizeof radreq->vector);
+	radreq->secret       = oldreq->secret;
+	radreq->request      = avl_dup(oldreq->request);
 
 	/* Proxy support fields */
-	authreq->realm         = dup_string(oldreq->realm);
-	authreq->server_ipaddr = oldreq->server_ipaddr;
-	authreq->server_id     = oldreq->server_id;
+	radreq->realm         = dup_string(oldreq->realm);
+	radreq->server_ipaddr = oldreq->server_ipaddr;
+	radreq->server_id     = oldreq->server_id;
 	
 	return 0;
 }
