@@ -1,22 +1,21 @@
 /* This file is part of GNU RADIUS.
- * Copyright (C) 2000,2001, Sergey Poznyakoff
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
- */
-#define RADIUS_MODULE 15
+   Copyright (C) 2000,2001, Sergey Poznyakoff
+  
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+  
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+  
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation,
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+
+#define RADIUS_MODULE_SNMPSERV_C
 #if defined(HAVE_CONFIG_H)
 # include <config.h>
 #endif
@@ -24,7 +23,7 @@
 #ifdef USE_SNMP
 
 #ifndef lint
-static char rcsid[] = "@(#) $Id: snmpserv.c,v 1.26 2001/04/17 06:42:09 gray Exp $";
+static char rcsid[] = "@(#) $Id: snmpserv.c,v 1.36 2001/09/06 09:58:16 gray Exp $";
 #endif
 
 #include <sys/types.h>
@@ -42,13 +41,16 @@ static char rcsid[] = "@(#) $Id: snmpserv.c,v 1.26 2001/04/17 06:42:09 gray Exp
 #include <radmibs.h>
 
 #define MAXOIDLEN 512
-
+ 
 struct snmp_pdu * snmp_agent_response(struct snmp_pdu *pdu, int access);
 
 int snmp_decode(SNMP_REQ *req, u_char *buf, int len);
 
 int variable_cmp(struct snmp_var *v1, struct snmp_var *v2);
 
+static NAS *nas_lookup_index(int ind);
+
+
 ACL *snmp_acl, *snmp_acl_tail;
 Community *commlist, *commlist_tail;
 Server_stat *server_stat;
@@ -167,6 +169,15 @@ int snmp_acct_handler(enum mib_node_cmd cmd, void *closure, subid_t subid,
 		      struct snmp_var **varp, int *errp);
 int snmp_acct_v_handler(enum mib_node_cmd cmd, void *closure, subid_t subid,
 			struct snmp_var **varp, int *errp);
+int snmp_serv_handler(enum mib_node_cmd cmd, void *closure, subid_t subid,
+		      struct snmp_var **varp, int *errp);
+int snmp_serv_queue_handler(enum mib_node_cmd cmd, void *closure,
+			    subid_t subid, struct snmp_var **varp, int *errp);
+int snmp_serv_mem_summary(enum mib_node_cmd cmd, void *closure,
+			    subid_t subid, struct snmp_var **varp, int *errp);
+int snmp_serv_class_handler(enum mib_node_cmd cmd, void *closure,
+			    subid_t subid, struct snmp_var **varp, int *errp);
+
 int snmp_stat_handler(enum mib_node_cmd cmd, void *closure, subid_t subid,
 		      struct snmp_var **varp, int *errp);
 
@@ -208,11 +219,21 @@ struct port_table_closure {
 	int port_index;
 };
 
-struct auth_mib_closure auth_closure, acct_closure;
-struct nas_closure nas_closure;
-struct nas_table nas_table;
-struct port_closure port_closure;
-struct port_table_closure port_table;
+struct queue_closure {
+	int queue_index;
+};
+
+struct mem_closure {
+	int mem_index;
+};
+
+static struct auth_mib_closure auth_closure, acct_closure;
+static struct nas_closure nas_closure;
+static struct nas_table nas_table;
+static struct port_closure port_closure;
+static struct port_table_closure port_table;
+static struct queue_closure queue_closure;
+static struct mem_closure mem_closure;
 
 static struct mib_data {
 	oid_t oid;
@@ -280,6 +301,32 @@ static struct mib_data {
 	oid_AccServNoRecords,                snmp_acct_v_handler,&acct_closure,
 	oid_AccServUnknownTypes,             snmp_acct_v_handler,&acct_closure,
 
+	/* Server */
+	oid_radiusServerUpTime,              snmp_serv_handler, NULL,
+	oid_radiusServerResetTime,           snmp_serv_handler, NULL,
+	oid_radiusServerState,               snmp_serv_handler, NULL,
+
+	/* Variable oids */
+	oid_queueIndex,        snmp_serv_queue_handler, &queue_closure,
+	oid_queueName,         snmp_serv_queue_handler, &queue_closure,	
+	oid_queueActive,       snmp_serv_queue_handler, &queue_closure,	
+	oid_queueHeld,         snmp_serv_queue_handler, &queue_closure,	
+	oid_queueTotal,        snmp_serv_queue_handler, &queue_closure,	
+
+	oid_memoryNumClasses,      snmp_serv_mem_summary,   NULL,
+	oid_memoryNumBuckets,      snmp_serv_mem_summary,   NULL,
+	oid_memoryBytesAllocated,  snmp_serv_mem_summary,   NULL,
+	oid_memoryBytesUsed,       snmp_serv_mem_summary,   NULL,
+	
+	oid_classIndex,	           snmp_serv_class_handler, &mem_closure,
+	oid_classSize,		   snmp_serv_class_handler, &mem_closure,
+	oid_classElsPerBucket,     snmp_serv_class_handler, &mem_closure,
+	oid_classNumBuckets,	   snmp_serv_class_handler, &mem_closure,
+	oid_classElsUsed,          snmp_serv_class_handler, &mem_closure,
+	
+	oid_memoryMallocBlocks,	   snmp_serv_mem_summary,   NULL,
+	oid_memoryMallocBytes,	   snmp_serv_mem_summary,   NULL,
+	
 	/* Statistics */
 	oid_StatIdent,                       snmp_stat_handler, NULL,
 	oid_StatUpTime,	                     snmp_stat_handler, NULL,
@@ -389,7 +436,34 @@ snmp_attach_nas_stat(nas)
 		nasstat->ipaddr = nas->ipaddr;
 	}
 	nasstat->index = server_stat->nas_index++;
-	nas->nas_stat = nasstat;
+	nas->app_data = nasstat;
+}
+
+static int
+nas_ip_cmp(a, b)
+	struct nas_stat **a, **b;
+{
+	return (*a)->ipaddr - (*b)->ipaddr;
+}
+
+void
+snmp_sort_nas_stat()
+{
+	struct nas_stat **nsarray, *nsp;
+	int i;
+
+	nsarray = emalloc(server_stat->nas_count*sizeof(nsarray[0]));
+	for (i = 0, nsp = (struct nas_stat*)(server_stat + 1); 
+		i < server_stat->nas_count; 
+		i++, nsp++) 
+		nsarray[i] = nsp;	
+	qsort(nsarray, server_stat->nas_count,
+	      sizeof(struct nas_stat*), nas_ip_cmp);
+	for (i = 0, nsp = (struct nas_stat*)(server_stat + 1); 
+		i < server_stat->nas_count; 
+		i++, nsp++) 
+		nsarray[i]->index = i+1;
+	efree(nsarray);
 }
 
 /* Mark reset of the auth server. Do not do any real work, though.
@@ -432,9 +506,9 @@ rad_snmp_respond(buf, len, sa)
 	req->sa = *sa;
 
 	debug(1,
-		("got %d bytes from %I",
+		("got %d bytes from %s",
 		 len,
-		 ntohl(req->sa.sin_addr.s_addr)));
+		 format_ipaddr(ntohl(req->sa.sin_addr.s_addr))));
 	
 	if (snmp_decode(req, buf, len)) {
 		free_entry(req);
@@ -474,10 +548,10 @@ snmp_decode(req, buf, len)
 	access = check_acl(req->sa.sin_addr.s_addr, comm);
 	if (!access) {
 		radlog(L_NOTICE,
-		       _("DENIED attempt to access community %s from %s (%I)"),
+		       _("DENIED attempt to access community %s from %s (%s)"),
 		       comm,
 		       ip_hostname(ntohl(req->sa.sin_addr.s_addr)),
-		       ntohl(req->sa.sin_addr.s_addr));
+		       format_ipaddr(ntohl(req->sa.sin_addr.s_addr)));
 		return 1;
 	}
 	req->pdu = pdu;
@@ -512,8 +586,8 @@ snmp_req_drop(type, req, status_str)
 	char *status_str;
 {
 	radlog(L_NOTICE,
-	       _("Dropping SNMP request from client %I: %s"),
-	       ntohl(req->sa.sin_addr.s_addr),
+	       _("Dropping SNMP request from client %s: %s"),
+	       format_ipaddr(ntohl(req->sa.sin_addr.s_addr)),
 	       status_str);
 }
 
@@ -527,6 +601,7 @@ snmp_answer(req, sock)
 	struct snmp_pdu *pdu;
 	int len;
 
+	log_open(L_SNMP);
 	pdu = snmp_agent_response(req->pdu, req->access);
 	if (pdu) {
 		session.version = SNMP_VERSION_1;
@@ -853,9 +928,7 @@ snmp_agent_response(pdu, access)
 
 			vresp = &answer->var;
 			/* Loop through all variables */
-			for (vpp = &pdu->var;
-			     *vpp;
-			     vpp = &(*vpp)->next) {
+			for (vpp = &pdu->var; *vpp; vpp = &(*vpp)->next) {
 				vp = *vpp;
 
 				index++;
@@ -918,6 +991,19 @@ timeval_diff(tva, tvb)
 		(tva->tv_usec - tvb->tv_usec)/10000;
 }
 
+serv_stat
+abridge_server_state()
+{
+	switch (server_stat->auth.status) {
+	case serv_init:
+	case serv_running:
+		return server_stat->auth.status;
+	case serv_other:
+	default:
+		return serv_other;
+	}
+}
+
 /* ************************************************************************* */
 /* Auth sub-tree */
 
@@ -994,13 +1080,14 @@ snmp_auth_var_get(subid, oid, errp)
 		gettimeofday(&tv, &tz);
 		ret->type = SMI_TIMETICKS;
 		ret->val_length = sizeof(counter);
-		ret->var_int = timeval_diff(&tv, &server_stat->auth.reset_time);
+		ret->var_int = timeval_diff(&tv,
+					    &server_stat->auth.reset_time);
 		break;
 
 	case MIB_KEY_AuthServConfigReset:
 		ret->type = ASN_INTEGER;
 		ret->val_length = sizeof(counter);
-		ret->var_int = server_stat->auth.status;
+		ret->var_int = abridge_server_state();
 		break;
 
 	case MIB_KEY_AuthServTotalAccessRequests:
@@ -1099,7 +1186,7 @@ snmp_auth_var_set(subid, vp, errp)
 		switch (subid) {
 			
 		case MIB_KEY_AccServConfigReset:
-			server_stat->acct.status = serv_init;
+			server_stat->auth.status = serv_init;
 			radlog(L_INFO,
 			     _("acct server re-initializing on SNMP request"));
 			break;
@@ -1164,7 +1251,7 @@ snmp_auth_var_next(subid, closure)
 	subid_t subid;
 	struct auth_mib_closure *closure;
 {
-	if (!findnasbyindex(subid)) 
+	if (!nas_lookup_index(subid)) 
 		return -1;
 
 	closure->nas_index = subid;
@@ -1197,8 +1284,8 @@ snmp_auth_var_v_get(subid, var, errp)
 	case MIB_KEY_AuthServBadAuthenticators:
 	case MIB_KEY_AuthServPacketsDropped:
 	case MIB_KEY_AuthServUnknownTypes:
-		if ((nas = findnasbyindex(subid)) != NULL &&
-		    nas->nas_stat) {
+		if ((nas = nas_lookup_index(subid)) != NULL &&
+		    nas->app_data) {
  			get_auth_nasstat(nas, ret, key);
 			break;
 		}
@@ -1217,18 +1304,20 @@ get_auth_nasstat(nas, var, key)
 	struct snmp_var *var;
 	int key;
 {
+	struct nas_stat *statp = nas->app_data;
+	
 	switch (key) {
 	case MIB_KEY_AuthClientIndex:
 		var->type = ASN_INTEGER;
 		var->val_length = sizeof(int);
-		var->var_int = nas->nas_stat->index;
+		var->var_int = statp->index;
 		break;
 		
 	case MIB_KEY_AuthClientAddress:
 		var->type = SMI_IPADDRESS;
 		var->val_length = sizeof(UINT4);
 		var->var_str = snmp_alloc(sizeof(UINT4));
-		*(UINT4*)var->var_str = ntohl(nas->nas_stat->ipaddr);
+		*(UINT4*)var->var_str = ntohl(statp->ipaddr);
 		break;
 
 	case MIB_KEY_AuthClientID:
@@ -1240,55 +1329,55 @@ get_auth_nasstat(nas, var, key)
 	case MIB_KEY_AuthServAccessRequests:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->auth.num_access_req;
+		var->var_int = statp->auth.num_access_req;
 		break;
 
 	case MIB_KEY_AuthServDupAccessRequests:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->auth.num_dup_req;
+		var->var_int = statp->auth.num_dup_req;
 		break;
 
 	case MIB_KEY_AuthServAccessAccepts:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->auth.num_accepts;
+		var->var_int = statp->auth.num_accepts;
 		break;
 
 	case MIB_KEY_AuthServAccessRejects:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->auth.num_rejects;
+		var->var_int = statp->auth.num_rejects;
 		break;
 
 	case MIB_KEY_AuthServAccessChallenges:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->auth.num_challenges;
+		var->var_int = statp->auth.num_challenges;
 		break;
 
 	case MIB_KEY_AuthServMalformedAccessRequests:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->auth.num_bad_req;
+		var->var_int = statp->auth.num_bad_req;
 		break;
 
 	case MIB_KEY_AuthServBadAuthenticators:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->auth.num_bad_auth;
+		var->var_int = statp->auth.num_bad_auth;
 		break;
 
 	case MIB_KEY_AuthServPacketsDropped:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->auth.num_dropped;
+		var->var_int = statp->auth.num_dropped;
 		break;
 
 	case MIB_KEY_AuthServUnknownTypes:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->auth.num_unknowntypes;
+		var->var_int = statp->auth.num_unknowntypes;
 		break;
 		
 	}
@@ -1376,7 +1465,7 @@ snmp_acct_var_get(subid, oid, errp)
 	case MIB_KEY_AccServConfigReset:
 		ret->type = ASN_INTEGER;
 		ret->val_length = sizeof(counter);
-		ret->var_int = server_stat->acct.status;
+		ret->var_int = abridge_server_state();
 		break;
 		
 	case MIB_KEY_AccServTotalRequests:
@@ -1552,8 +1641,8 @@ snmp_acct_var_v_get(subid, var, errp)
 	case MIB_KEY_AccServMalformedRequests:
 	case MIB_KEY_AccServNoRecords:
 	case MIB_KEY_AccServUnknownTypes:         
-		if ((nas = findnasbyindex(subid)) != NULL &&
-		     nas->nas_stat) {
+		if ((nas = nas_lookup_index(subid)) != NULL &&
+		     nas->app_data) {
  			get_acct_nasstat(nas, ret, key);
 			break;
 		}
@@ -1572,18 +1661,20 @@ get_acct_nasstat(nas, var, key)
 	struct snmp_var *var;
 	int key;
 {
+	struct nas_stat *statp = nas->app_data;
+	
 	switch (key) {
 	case MIB_KEY_AccClientIndex:
 		var->type = ASN_INTEGER;
 		var->val_length = sizeof(int);
-		var->var_int = nas->nas_stat->index;
+		var->var_int = statp->index;
 		break;
 
 	case MIB_KEY_AccClientAddress:
 		var->type = SMI_IPADDRESS;
 		var->val_length = sizeof(UINT4);
 		var->var_str = snmp_alloc(sizeof(UINT4));
-		*(UINT4*)var->var_str = ntohl(nas->nas_stat->ipaddr);
+		*(UINT4*)var->var_str = ntohl(statp->ipaddr);
 		break;
 
 	case MIB_KEY_AccClientID:
@@ -1595,49 +1686,586 @@ get_acct_nasstat(nas, var, key)
 	case MIB_KEY_AccServPacketsDropped:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->acct.num_dropped;
+		var->var_int = statp->acct.num_dropped;
 		break;
 
 	case MIB_KEY_AccServRequests:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->acct.num_req;
+		var->var_int = statp->acct.num_req;
 		break;
 
 	case MIB_KEY_AccServDupRequests:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->acct.num_dup_req;
+		var->var_int = statp->acct.num_dup_req;
 		break;
 
 	case MIB_KEY_AccServResponses:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->acct.num_resp;
+		var->var_int = statp->acct.num_resp;
 		break;
 
 	case MIB_KEY_AccServBadAuthenticators:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->acct.num_bad_sign;
+		var->var_int = statp->acct.num_bad_sign;
 		break;
 
 	case MIB_KEY_AccServMalformedRequests:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->acct.num_bad_req;
+		var->var_int = statp->acct.num_bad_req;
 		break;
 
 	case MIB_KEY_AccServNoRecords:
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->acct.num_norecords;
+		var->var_int = statp->acct.num_norecords;
 		break;
 
 	case MIB_KEY_AccServUnknownTypes:         
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->acct.num_unknowntypes;
+		var->var_int = statp->acct.num_unknowntypes;
+		break;
+	}
+}
+
+/* ************************************************************************* */
+/* Server */
+struct snmp_var *snmp_serv_var_get(subid_t subid, oid_t oid, int *errp);
+int snmp_serv_var_set(subid_t subid, struct snmp_var **vp, int *errp);
+
+/* Handler function for fixed oids from the server subtree */
+/*ARGSUSED*/
+int
+snmp_serv_handler(cmd, closure, subid, varp, errp)
+	enum mib_node_cmd cmd;
+	void *closure;
+	subid_t subid;
+	struct snmp_var **varp;
+	int *errp;
+{
+	oid_t oid = (*varp)->name;
+	
+	switch (cmd) {
+	case MIB_NODE_GET:
+		if ((*varp = snmp_serv_var_get(subid, oid, errp)) == NULL)
+			return -1;
+		break;
+		
+	case MIB_NODE_SET:
+		return snmp_serv_var_set(subid, varp, errp);
+		
+	case MIB_NODE_SET_TRY:
+		return snmp_serv_var_set(subid, varp, errp);
+		
+	case MIB_NODE_RESET:
+		break;
+		
+	default: /* unused: should never get there */
+		abort();
+
+	}
+	
+	return 0;
+}
+
+struct snmp_var *
+snmp_serv_var_get(subid, oid, errp)
+	subid_t subid;
+	oid_t oid;
+	int *errp;
+{
+	struct snmp_var *ret;
+	struct timeval tv;
+	struct timezone tz;
+	char *p;
+	
+	ret = snmp_var_create(oid);
+	*errp = SNMP_ERR_NOERROR;
+
+	switch (subid) {
+
+	case MIB_KEY_radiusServerUpTime:
+		gettimeofday(&tv, &tz);
+		ret->type = SMI_TIMETICKS;
+		ret->val_length = sizeof(counter);
+		ret->var_int = timeval_diff(&tv, &server_stat->start_time);
+		break;
+
+	case MIB_KEY_radiusServerResetTime:
+		gettimeofday(&tv, &tz);
+		ret->type = SMI_TIMETICKS;
+		ret->val_length = sizeof(counter);
+		ret->var_int = timeval_diff(&tv,
+					    &server_stat->auth.reset_time);
+		break;
+
+	case MIB_KEY_radiusServerState:
+		ret->type = ASN_INTEGER;
+		ret->val_length = sizeof(counter);
+		ret->var_int = server_stat->auth.status;/*FIXME*/
+		break;
+	default:
+		*errp = SNMP_ERR_NOSUCHNAME;
+		snmp_var_free(ret);
+		return NULL;
+	}
+	return ret;
+}
+
+int
+snmp_serv_var_set(subid, vp, errp)
+	subid_t subid;
+	struct snmp_var **vp;
+	int *errp;
+{
+	if (errp) { /* just test */
+		*errp = SNMP_ERR_NOERROR;
+		switch (subid) {
+		
+		case MIB_KEY_radiusServerState:
+			if ((*vp)->type != ASN_INTEGER) {
+				*errp = SNMP_ERR_BADVALUE;
+				*vp = NULL;
+			} else {
+				switch ((*vp)->var_int) {
+				case serv_reset:
+				case serv_init:
+				case serv_running:
+				case serv_suspended:
+				case serv_shutdown:
+					break;
+				default:
+					*errp = SNMP_ERR_BADVALUE;
+					*vp = NULL;
+				}
+			}
+			break;
+		default:
+			*errp = SNMP_ERR_BADVALUE;
+			(*vp) = NULL;
+		}
+	} else {
+		/* do set it */
+		*vp = snmp_var_dup(*vp);
+		
+		switch (subid) {
+
+		case MIB_KEY_radiusServerState:
+			server_stat->auth.status = (*vp)->var_int;
+			switch ((*vp)->var_int) {
+			case serv_reset:
+				radlog(L_NOTICE,
+				       _("server re-initializing on SNMP request"));
+				break;
+			case serv_init:
+				radlog(L_NOTICE,
+				       _("server restart on SNMP request"));
+				break;
+			case serv_running:
+				radlog(L_NOTICE,
+				       _("server continuing on SNMP request"));
+				break;
+			case serv_suspended:
+				radlog(L_NOTICE,
+				       _("server suspending on SNMP request"));
+				break;
+			case serv_shutdown:
+				radlog(L_NOTICE,
+				       _("server shutting down on SNMP request"));
+				break;
+			}
+			break;
+
+		}
+	}
+	return (*vp == NULL);
+}
+
+
+/* Variable oids */
+
+/* Queue table */
+static struct snmp_var *snmp_queue_get(subid_t subid, struct snmp_var *var,
+				       int *errp);
+static void get_queue_stat(int qno, struct snmp_var *var, subid_t key);
+
+int
+snmp_serv_queue_handler(cmd, closure, subid, varp, errp)
+	enum mib_node_cmd cmd;
+	void *closure;
+	subid_t subid;
+	struct snmp_var **varp;
+	int *errp;
+{
+	struct queue_closure *p = (struct queue_closure*)closure;
+	
+	switch (cmd) {
+	case MIB_NODE_GET:
+		if ((*varp = snmp_queue_get(subid, *varp, errp)) == NULL)
+			return -1;
+		break;
+		
+	case MIB_NODE_SET:
+	case MIB_NODE_SET_TRY:
+		/* None of these can be set */
+		if (errp)
+			*errp = SNMP_ERR_NOSUCHNAME;
+		return -1;
+		
+	case MIB_NODE_COMPARE: 
+		return 0;
+		
+	case MIB_NODE_NEXT:
+		if (subid < R_MAX) {
+			p->queue_index = subid+1;
+			return 0;
+		}
+		return -1;
+		
+	case MIB_NODE_GET_SUBID:
+		return p->queue_index;
+		
+	case MIB_NODE_RESET:
+		p->queue_index = 1;
+		break;
+
+	}
+	
+	return 0;
+}
+
+struct snmp_var *
+snmp_queue_get(subid, var, errp)
+	subid_t subid;
+	struct snmp_var *var;
+	int *errp;
+{
+	struct snmp_var *ret;
+	subid_t key;
+	oid_t oid = var->name;
+	
+	ret = snmp_var_create(oid);
+	*errp = SNMP_ERR_NOERROR;
+
+	switch (key = SUBID(oid, OIDLEN(oid)-2)) {
+
+	case MIB_KEY_queueIndex:
+	case MIB_KEY_queueName:			
+	case MIB_KEY_queueActive:
+	case MIB_KEY_queueHeld:
+	case MIB_KEY_queueTotal:
+		if (subid-1 < R_MAX) {
+			get_queue_stat(subid-1, ret, key);
+			break;
+
+		}
+		/*FALLTHRU*/
+		
+	default:
+		*errp = SNMP_ERR_NOSUCHNAME;
+		snmp_var_free(ret);
+		return NULL;
+	}
+	return ret;
+}
+
+void
+get_queue_stat(qno, var, key)
+	int qno;
+	struct snmp_var *var;
+	subid_t key;
+{
+	struct timeval tv;
+	struct timezone tz;
+	QUEUE_STAT stat;
+	
+	stat_request_list(stat);
+	switch (key) {
+	case MIB_KEY_queueIndex:
+		var->type = ASN_INTEGER;
+		var->val_length = sizeof(int);
+		var->var_int = qno+1;
+		break;
+		
+	case MIB_KEY_queueName:
+		var->type = ASN_OCTET_STR;
+		var->val_length = strlen(request_class[qno].name);
+		var->var_str = snmp_strdup(request_class[qno].name);
+		break;
+
+	case MIB_KEY_queueActive:
+		var->type = SMI_COUNTER32;
+		var->val_length = sizeof(counter);
+		var->var_int = stat[qno][0];
+		break;
+		
+	case MIB_KEY_queueHeld:
+		var->type = SMI_COUNTER32;
+		var->val_length = sizeof(counter);
+		var->var_int = stat[qno][1];
+		break;
+
+	case MIB_KEY_queueTotal:
+		var->type = SMI_COUNTER32;
+		var->val_length = sizeof(counter);
+		var->var_int = stat[qno][0]+stat[qno][1];
+		break;
+	}
+}
+
+/* Memory table */
+static struct snmp_var *snmp_mem_get(subid_t subid, oid_t oid, int *errp);
+static void get_mem_stat(int no, struct snmp_var *var, subid_t key);
+static struct snmp_var *snmp_class_get(subid_t subid,
+				       struct snmp_var *varp, int *errp);
+static int _mem_get_class(subid_t, CLASS_STAT*);
+static void get_class_stat(CLASS_STAT *stat, struct snmp_var *var,
+			   subid_t key);
+
+int
+snmp_serv_mem_summary(cmd, closure, subid, varp, errp)
+	enum mib_node_cmd cmd;
+	void *closure;
+	subid_t subid;
+	struct snmp_var **varp;
+	int *errp;
+{
+	oid_t oid = (*varp)->name;
+	CLASS_STAT stat;
+	
+	switch (cmd) {
+	case MIB_NODE_GET:
+		if ((*varp = snmp_mem_get(subid, oid, errp)) == NULL)
+			return -1;
+		break;
+		
+	case MIB_NODE_SET:
+	case MIB_NODE_SET_TRY:
+		/* None of these can be set */
+		if (errp)
+			*errp = SNMP_ERR_NOSUCHNAME;
+		return -1;
+		
+	case MIB_NODE_RESET:
+		break;
+
+	default:
+		abort();
+
+	}
+	
+	return 0;
+}
+
+struct snmp_var *
+snmp_mem_get(subid, oid, errp)
+	subid_t subid;
+	oid_t oid;
+	int *errp;
+{
+	struct snmp_var *ret;
+	struct timeval tv;
+	struct timezone tz;
+	MEM_STAT stat;
+	
+	ret = snmp_var_create(oid);
+	*errp = SNMP_ERR_NOERROR;
+
+	mem_get_stat(&stat);
+	
+	switch (subid) {
+
+	case MIB_KEY_memoryNumClasses:
+		ret->type = SMI_COUNTER32;
+		ret->val_length = sizeof(counter);
+		ret->var_int = stat.class_cnt;
+		break;
+		
+	case MIB_KEY_memoryNumBuckets:
+		ret->type = SMI_COUNTER32;
+		ret->val_length = sizeof(counter);
+		ret->var_int = stat.bucket_cnt;
+		break;
+
+	case MIB_KEY_memoryBytesAllocated:
+		ret->type = SMI_COUNTER32;
+		ret->val_length = sizeof(counter);
+		ret->var_int = stat.bytes_allocated;
+		break;
+
+	case MIB_KEY_memoryBytesUsed:
+		ret->type = SMI_COUNTER32;
+		ret->val_length = sizeof(counter);
+		ret->var_int = stat.bytes_used;
+		break;
+		
+	case MIB_KEY_memoryMallocBlocks:
+		ret->type = SMI_COUNTER32;
+		ret->val_length = sizeof(counter);
+#ifdef LEAK_DETECTOR
+		ret->var_int = mallocstat.count;
+#else
+		ret->var_int = 0;
+#endif
+		break;
+		
+	case MIB_KEY_memoryMallocBytes:
+		ret->type = SMI_COUNTER32;
+		ret->val_length = sizeof(counter);
+#ifdef LEAK_DETECTOR
+		ret->var_int = mallocstat.size;
+#else
+		ret->var_int = 0;
+#endif
+		break;
+		
+	default:
+		*errp = SNMP_ERR_NOSUCHNAME;
+		snmp_var_free(ret);
+		return NULL;
+	}
+	return ret;
+}
+
+static int
+class_counter(stat, ret)
+	CLASS_STAT *stat;
+	CLASS_STAT *ret;
+{
+	if (stat->index == ret->index) {
+		*ret = *stat;
+		return 1;
+	}
+	return 0;
+}
+
+int
+_mem_get_class(subid, stat)
+	subid_t subid;
+	CLASS_STAT *stat;
+{
+	stat->index = subid;
+	return mem_stat_enumerate(class_counter, stat);
+}
+
+int
+snmp_serv_class_handler(cmd, closure, subid, varp, errp)
+	enum mib_node_cmd cmd;
+	void *closure;
+	subid_t subid;
+	struct snmp_var **varp;
+	int *errp;
+{
+	struct mem_closure *p = (struct mem_closure*)closure;
+	CLASS_STAT stat;
+	
+	switch (cmd) {
+	case MIB_NODE_GET:
+		if ((*varp = snmp_class_get(subid, *varp, errp)) == NULL)
+			return -1;
+		break;
+		
+	case MIB_NODE_SET:
+	case MIB_NODE_SET_TRY:
+		/* None of these can be set */
+		if (errp)
+			*errp = SNMP_ERR_NOSUCHNAME;
+		return -1;
+		
+	case MIB_NODE_COMPARE: 
+		return 0;
+		
+	case MIB_NODE_NEXT:
+		if (_mem_get_class(subid, &stat)) {
+			p->mem_index = subid+1;
+			return 0;
+		}
+		return -1;
+		
+	case MIB_NODE_GET_SUBID:
+		return p->mem_index;
+		
+	case MIB_NODE_RESET:
+		p->mem_index = 1;
+		break;
+
+	}
+	
+	return 0;
+}
+
+struct snmp_var *
+snmp_class_get(subid, var, errp)
+	subid_t subid;
+	struct snmp_var *var;
+	int *errp;
+{
+	struct snmp_var *ret;
+	subid_t key;
+	oid_t oid = var->name;
+	CLASS_STAT stat;
+	
+	ret = snmp_var_create(oid);
+	*errp = SNMP_ERR_NOERROR;
+
+	switch (key = SUBID(oid, OIDLEN(oid)-2)) {
+
+	case MIB_KEY_classIndex:
+	case MIB_KEY_classSize:			
+	case MIB_KEY_classElsPerBucket:
+	case MIB_KEY_classNumBuckets:
+	case MIB_KEY_classElsUsed:
+		if (_mem_get_class(subid-1, &stat)) {
+			get_class_stat(&stat, ret, key);
+			break;
+
+		}
+		/*FALLTHRU*/
+		
+	default:
+		*errp = SNMP_ERR_NOSUCHNAME;
+		snmp_var_free(ret);
+		return NULL;
+	}
+	return ret;
+}
+
+void
+get_class_stat(stat, var, key)
+	CLASS_STAT *stat;
+	struct snmp_var *var;
+	subid_t key;
+{
+	switch (key) {
+	case MIB_KEY_classIndex:
+		var->type = ASN_INTEGER;
+		var->val_length = sizeof(int);
+		var->var_int = stat->index+1;
+		break;
+	case MIB_KEY_classSize:
+		var->type = ASN_INTEGER;
+		var->val_length = sizeof(int);
+		var->var_int = stat->elsize;
+		break;
+	case MIB_KEY_classElsPerBucket:
+		var->type = SMI_COUNTER32;
+		var->val_length = sizeof(counter);
+		var->var_int = stat->elcnt;
+		break;
+	case MIB_KEY_classNumBuckets:
+		var->type = SMI_COUNTER32;
+		var->val_length = sizeof(counter);
+		var->var_int = stat->bucket_cnt;
+		break;
+	case MIB_KEY_classElsUsed:
+		var->type = SMI_COUNTER32;
+		var->val_length = sizeof(counter);
+		var->var_int = stat->allocated_cnt;
 		break;
 	}
 }
@@ -1808,11 +2436,12 @@ snmp_stat_nas(num, cmd, closure, subid, varp, errp)
 			(closure->quad[2]<<8) +
 			closure->quad[3];
 
-		if ((nas = nas_find(ip)) == NULL) {
+		if ((nas = nas_lookup_ip(ip)) == NULL) {
 			return -1;
 		}
-
-		if ((nas = findnasbyindex(nas->nas_stat->index+1)) == NULL) {
+		
+		nsp = nas->app_data;
+		if ((nas = nas_lookup_index(nsp->index+1)) == NULL) {
 			return -1;
 		}
 
@@ -1827,7 +2456,7 @@ snmp_stat_nas(num, cmd, closure, subid, varp, errp)
 		
 	case MIB_NODE_RESET:
 		if (num == 0) {
-			if (nas = findnasbyindex(1))
+			if (nas = nas_lookup_index(1))
 				for (num = 0; num < 4; num++)
 					closure->quad[num] = (nas->ipaddr >>
 							      (8*(3-num))) & 0xff;
@@ -1911,7 +2540,7 @@ snmp_nas_table(cmd, closure, subid, varp, errp)
 		return -1;
 		
 	case MIB_NODE_NEXT:
-		if (!findnasbyindex(subid+1))
+		if (!nas_lookup_index(subid+1))
 			return -1;
 		((struct nas_table*)closure)->row = subid+1;
 		break;
@@ -1955,8 +2584,7 @@ snmp_nas_table_get(subid, oid, errp)
 	case MIB_KEY_NASLines:
 	case MIB_KEY_NASLinesInUse:
 	case MIB_KEY_NASLinesIdle:
-		if ((nas = findnasbyindex(subid)) != NULL &&
-		     nas->nas_stat) {
+		if ((nas = nas_lookup_index(subid)) != NULL && nas->app_data) {
  			get_stat_nasstat(nas, ret, key);
 			break;
 		}
@@ -1975,12 +2603,14 @@ get_stat_nasstat(nas, var, ind)
 	struct snmp_var *var;
 	int ind;
 {
+	struct nas_stat *statp = nas->app_data;
+	
 	switch (ind) {
 	case MIB_KEY_NASAddress:
 		var->type = SMI_IPADDRESS;
 		var->val_length = sizeof(UINT4);
 		var->var_str = snmp_alloc(sizeof(UINT4));
-		*(UINT4*)var->var_str = ntohl(nas->nas_stat->ipaddr);
+		*(UINT4*)var->var_str = ntohl(statp->ipaddr);
 		break;
 
 	case MIB_KEY_NASID:
@@ -1993,22 +2623,22 @@ get_stat_nasstat(nas, var, ind)
 		stat_count_ports();
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->ports_active +
-			       nas->nas_stat->ports_idle;
+		var->var_int = statp->ports_active +
+			       statp->ports_idle;
 		break;
 
 	case MIB_KEY_NASLinesInUse:
 		stat_count_ports();
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->ports_active;
+		var->var_int = statp->ports_active;
 		break;
 
 	case MIB_KEY_NASLinesIdle:
 		stat_count_ports();
 		var->type = SMI_COUNTER32;
 		var->val_length = sizeof(counter);
-		var->var_int = nas->nas_stat->ports_idle;
+		var->var_int = statp->ports_idle;
 		break;
 
 	}
@@ -2050,7 +2680,7 @@ snmp_port_index1(cmd, closure, subid, varp, errp)
 		
 	case MIB_NODE_RESET:
 		pind->nas_index = 1;
-		while ((nas = findnasbyindex(pind->nas_index)) && 
+		while ((nas = nas_lookup_index(pind->nas_index)) && 
 		       (pind->port_no = stat_get_next_port_no(nas, 0)) == 0)
 			pind->nas_index++;
 		break; 
@@ -2074,7 +2704,7 @@ snmp_port_index2(cmd, closure, subid, varp, errp)
 	
 	switch (cmd) {
 	case MIB_NODE_GET:
-		if ((nas = findnasbyindex(pind->nas_index)) == NULL ||
+		if ((nas = nas_lookup_index(pind->nas_index)) == NULL ||
 		    (index = stat_get_port_index(nas, pind->port_no)) == 0) {
 			*errp = SNMP_ERR_NOSUCHNAME;
 			return -1;
@@ -2099,7 +2729,7 @@ snmp_port_index2(cmd, closure, subid, varp, errp)
 		return 0;
 		
 	case MIB_NODE_NEXT:
-		if ((nas = findnasbyindex(pind->nas_index)) == NULL)
+		if ((nas = nas_lookup_index(pind->nas_index)) == NULL)
 			return -1;
 		index = stat_get_next_port_no(nas, pind->port_no);
 		if (index > 0) {
@@ -2107,7 +2737,7 @@ snmp_port_index2(cmd, closure, subid, varp, errp)
 			break;
 		}
 		/* move to next nas */
-		while ((nas = findnasbyindex(++pind->nas_index)) && 
+		while ((nas = nas_lookup_index(++pind->nas_index)) && 
 		       (pind->port_no = stat_get_next_port_no(nas, 0)) == 0)
 			;
 
@@ -2235,12 +2865,13 @@ get_port_stat(port, var, key)
 	switch (key) {
 
 	case MIB_KEY_StatPortNASIndex:
-		nas = nas_find(port->ip);
+		nas = nas_lookup_ip(port->ip);
 		var->type = ASN_INTEGER;
 		var->val_length = sizeof(counter);
-		if (nas)
-			var->var_int = nas->nas_stat->index;
-		else
+		if (nas && nas->app_data) {
+			struct nas_stat *nsp = nas->app_data;
+			var->var_int = nsp->index;
+		} else
 			var->var_int = 0;
 		break;
 
@@ -2338,6 +2969,21 @@ get_port_stat(port, var, key)
 	}
 }
 
+NAS *
+nas_lookup_index(ind)
+	int ind;
+{
+	NAS *nas;
+	struct nas_stat *ns;
+	
+	for (nas = nas_next(NULL); nas; nas = nas_next(nas)) {
+		ns = nas->app_data;
+		if (ns && ns->index == ind)
+			break;
+	}
+	return nas;
+}
+
 
 #endif
 
