@@ -1,26 +1,24 @@
 /* This file is part of GNU RADIUS.
- * Copyright (C) 2000,2001, Sergey Poznyakoff
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
- */
+   Copyright (C) 2000,2001, Sergey Poznyakoff
+  
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+  
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+  
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation,
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 
-#define RADIUS_MODULE 9
+#define RADIUS_MODULE_SQL_C
 #ifndef lint
 static char rcsid[] =
-"@(#) $Id: sql.c,v 1.12 2001/04/11 07:58:38 gray Exp $";
+"@(#) $Id: sql.c,v 1.29 2001/12/27 13:06:03 gray Exp $";
 #endif
 
 #if defined(HAVE_CONFIG_H)
@@ -41,8 +39,9 @@ static char rcsid[] =
 
 static void sql_check_config(SQL_cfg *);
 static struct sql_connection *create_sql_connection(int type);
-static struct sql_connection *attach_sql_connection(int type, qid_t qid);
-static void unattach_sql_connection(int type, qid_t  qid);
+static struct sql_connection *attach_sql_connection(int type, qid_t qid, 
+                                                    int *keepopen);
+static void unattach_sql_connection(int type, qid_t qid, int keepopen);
 static void free_sql_connection(int type, qid_t qid);
 static void close_sql_connections(int type);
 
@@ -77,11 +76,15 @@ SQL_cfg sql_cfg;
 #define STMT_MAX_AUTH_CONNECTIONS  18
 #define STMT_MAX_ACCT_CONNECTIONS  19
 #define STMT_GROUP_QUERY           20
+#define STMT_REPLY_ATTR_QUERY            21
+#define STMT_INTERFACE             22
+#define STMT_CHECK_ATTR_QUERY      23
 
 static FILE  *sqlfd;
 static int line_no;
 static char *cur_line, *cur_ptr;
 static struct obstack stack;
+static int obstack_ready;
 static int stmt_type;
 
 struct keyword sql_keyword[] = {
@@ -99,13 +102,16 @@ struct keyword sql_keyword[] = {
 	"acct_db",            STMT_ACCT_DB,
 	"auth_query",         STMT_AUTH_QUERY,
 	"group_query",        STMT_GROUP_QUERY,
+	"attr_query",         STMT_REPLY_ATTR_QUERY,
+	"reply_attr_query",   STMT_REPLY_ATTR_QUERY,
+	"check_attr_query",   STMT_CHECK_ATTR_QUERY,
 	"acct_start_query",   STMT_ACCT_START_QUERY,
 	"acct_stop_query",    STMT_ACCT_STOP_QUERY,
 	"acct_alive_query",   STMT_ACCT_KEEPALIVE_QUERY,
 	"acct_keepalive_query", STMT_ACCT_KEEPALIVE_QUERY,
 	"acct_nasup_query",   STMT_ACCT_NASUP_QUERY,
 	"acct_nasdown_query", STMT_ACCT_NASDOWN_QUERY,
-	"query_buffer_size",  STMT_QUERY_BUFFER_SIZE,
+	"interface",          STMT_INTERFACE,
 	NULL,
 };
 
@@ -228,7 +234,7 @@ sql_digest(cfg)
 #define STRLEN(a) (a ? strlen(a) : 0)
 #define STPCPY(d,s) if (s) { strcpy(d,s); d += strlen(s); }
 	
-	length =  5 * sizeof(int)
+	length =  6 * sizeof(int)
 		+ 2 * sizeof(unsigned)
 		+ STRLEN(cfg->server)
 		+ STRLEN(cfg->login)
@@ -242,6 +248,9 @@ sql_digest(cfg)
 	
 	*(int*)p = length;
 	p += sizeof(int);
+
+	*(int*)p = cfg->interface;
+	p += sizeof(int);
 	
 	*(int*)p = cfg->keepopen;
 	p += sizeof(int);
@@ -306,11 +315,10 @@ rad_sql_init()
 	size_t bufsize = 0;
 	time_t timeout;
 	SQL_cfg new_cfg;
-	
+
 #define FREE(a) if (a) efree(a); a = NULL
 
 	bzero(&new_cfg, sizeof(new_cfg));
-	new_cfg.buf.size = 1024; 
 	new_cfg.keepopen = 0;
 	new_cfg.idle_timeout = 4*3600; /* four hours */
 	new_cfg.max_connections[SQL_AUTH] = 16;
@@ -327,7 +335,10 @@ rad_sql_init()
 	}
 	line_no = 0;
 	cur_line = NULL;
-	obstack_init(&stack);
+	if (!obstack_ready) {
+		obstack_init(&stack);
+		obstack_ready = 1;
+	} 
 	while (getline()) {
 		if (stmt_type == -1) {
 			radlog(L_ERR,
@@ -456,18 +467,33 @@ rad_sql_init()
 			new_cfg.acct_nasdown_query = estrdup(cur_ptr);
 			break;
 			
+		case STMT_REPLY_ATTR_QUERY:
+			new_cfg.reply_attr_query = estrdup(cur_ptr);
+			break;
+
+		case STMT_CHECK_ATTR_QUERY:
+			new_cfg.check_attr_query = estrdup(cur_ptr);
+			break;
+			
 		case STMT_QUERY_BUFFER_SIZE:
-			bufsize = strtol(cur_ptr, &ptr, 0);
-			if (*ptr != 0 && !isspace(*ptr)) {
-				radlog(L_ERR, _("%s:%d: number parse error"),
+			radlog(L_WARN, "%s:%d: query_buffer_size is obsolete",
+				       sqlfile, line_no);
+			break;
+			
+		case STMT_INTERFACE:
+			new_cfg.interface = disp_sql_interface_index(cur_ptr);
+			if (!new_cfg.interface) {
+				radlog(L_WARN, "%s:%d: Unsupported SQL interface.",
 				       sqlfile, line_no);
-				bufsize = 0;
 			}
 			break;
 		}
 		
 	}
-	obstack_free(&stack, NULL);
+
+	if (cur_line)
+		obstack_free(&stack, cur_line);
+//	obstack_free(&stack,NULL);
 	fclose(sqlfd);
 	efree(sqlfile);
 
@@ -489,11 +515,11 @@ rad_sql_init()
 	FREE(sql_cfg.acct_nasup_query);
 	FREE(sql_cfg.acct_nasdown_query);
 	FREE(sql_cfg.acct_keepalive_query);
-	FREE(sql_cfg.buf.ptr);
-
+	FREE(sql_cfg.reply_attr_query);
+	FREE(sql_cfg.check_attr_query);
+	
 	/* copy new config */
 	sql_cfg = new_cfg;
-	alloc_buffer(&sql_cfg.buf, bufsize);
 		
 	return 0;
 }
@@ -559,9 +585,6 @@ sql_check_config(cfg)
 		FREE_IF_EMPTY(cfg->acct_nasup_query);
 	}
 	
-	if (cfg->port == 0)
-		cfg->port = RAD_SQL_PORT;
-	
 	debug(1, ("SQL init using: %s:%d,%s,%s,%s,%d,%ld,%d,%d",
 	       cfg->server,
 	       cfg->port,
@@ -574,6 +597,17 @@ sql_check_config(cfg)
 	       cfg->doauth));
 }
 
+static qid_t last_qid = 0;
+
+qid_t
+rad_sql_get_qid()
+{
+	/* Reserve 0 for debugging purposes */
+	if (++last_qid == 0)
+		++last_qid;
+	return last_qid;
+}
+
 /* ************************************************************************* */
 /* Internal routines
  *
@@ -652,7 +686,7 @@ create_sql_connection(type)
 		conn_last = conn;
 
 	conn_count[type]++;
-	rad_sql_reconnect(type, conn);	
+	disp_sql_reconnect(sql_cfg.interface, type, conn);	
 	return conn;
 }
 
@@ -670,7 +704,7 @@ close_sql_connection(conn, prev)
 		 conn_count[SQL_AUTH] + conn_count[SQL_ACCT] - 1));
 	
 	if (conn->connected)
-		rad_sql_disconnect(conn);
+		disp_sql_disconnect(sql_cfg.interface, conn);
 	if (prev)
 		prev->next = conn->next;
 	if (conn_first == conn)
@@ -689,56 +723,93 @@ close_sql_connection(conn, prev)
  *          timestamp and return.
  *       Additionally, while scanning, close all connections which have been
  *       idle for more than sql_cfg.idle_timeout seconds.
+ *
+ * Modified by kaz:
+ *       Remove the code to scan for idle connections.  
+ *
+ *       Why?  This function is only called when the RADIUS server is
+ *       about to log something to the SQL server. It is called in two
+ *       places: in the master_process when it calls request_setup and 
+ *       when the child process calls rad_accounting/rad_authenticate. 
+ *
+ *       Whats the problem?  If the RADIUS server is configured to 
+ *       "keepopen" the SQL connections, then there is a potential 
+ *       that the RADIUS server may be idle for a time greater than 
+ *       the time expected by the SQL server's idle timeout, which 
+ *       could result in an error.  On MySQL, it will complain about 
+ *       an "Aborted connection ## to db".  For example, my RADIUS 
+ *       server was using "keepopen", but it was not very busy.  My
+ *       database "MySQL" was configured with a default "wait_timeout"
+ *       of 8 hours.  Everytime, the 8 hours were exceeded, MySQL would
+ *       mark the connection dead, yet, the RADIUS server still thought
+ *       the connection was open and it resulted in error messages in 
+ *       my MySQL logs.
+ *
+ *       Whats the solution?  Move the idle checking to the main loop
+ *       of radiusd.c.  rad_select should be modified to wait only for 
+ *       a fixed interval, perhaps 60 seconds, instead of blocking 
+ *       indefinitely.  A new function should also be created in sql.c
+ *       called rad_sql_idle_check() which will close connections that
+ *       are now idle.
+ * gray: The connection stuff is way too complicated and ugly. It must
+ *       be reritten from scratch. See `NOTE' below.
  */
 struct sql_connection *
-attach_sql_connection(type, qid)
+attach_sql_connection(type, qid, keepopen)
 	int type;
 	qid_t qid;
+	int *keepopen;
 {
 	struct sql_connection *conn = NULL, *prev, *next;
 	time_t now = time(NULL);
 
-	debug(1, ("attaching %d,%d", type, qid));
-
 	prev = NULL;
-	if (sql_cfg.keepopen) {
+	if (qid == 0) {
+		qid = rad_sql_get_qid();
+		debug(1, ("creating %d,%d", type, qid));
 		print_queue();
+		*keepopen = 0;
+	} else {
+		debug(1, ("attaching %d,%d", type, qid));
 
-		conn = conn_first;
-		while (conn) {
-			if (conn->type == type) { 
-				if (conn->qid == qid)
-					return conn;
-				if (conn->qid == 0)
-					break;
-			}
-			next = conn->next;
-			if (master_process() &&
-			    conn->qid == 0 &&
-			    (now - conn->last_used) >= sql_cfg.idle_timeout) {
-				/* Close the connection */
-				debug(1,
-			     ("connection reached idle timeout: %p,%d",
-			      conn, conn->type));
-				close_sql_connection(conn, prev);
-			} else
+		*keepopen = sql_cfg.keepopen;
+		if (sql_cfg.keepopen) {
+			print_queue();
+
+			conn = conn_first;
+			while (conn) {
+				if (conn->type == type) { 
+					if (conn->qid == qid)
+						return conn;
+					if (conn->qid == 0)
+						break;
+				}
+				next = conn->next;
 				prev = conn;
-		
-			conn = next;
+				conn = next;
+			}
 		}
 	}
-	
 	if (!conn) {
-
-		insist(conn_last == prev);
 		
-		if (conn_count[type] >= sql_cfg.max_connections[type]) {
-			radlog(L_CRIT,
+		if (prev) {
+			insist(conn_last == prev);
+		
+			if (conn_count[type] >= sql_cfg.max_connections[type]) {
+				radlog(L_CRIT,
 	 _("can't create new %s SQL connection: too many connections open"),
-			       service_name[type]);
-			return NULL;
+			       	       service_name[type]);
+				return NULL;
+			}
 		}
-		
+		/* NOTE: prev == NULL (i.e. input qid==0) violates 
+ 		   conn_count[type]. It is possible in two cases:
+			1. Where sql_cfg.keepopen==1. 
+			   The opened connection will be closed immediately
+			   after execution of the query.	
+			2. In master process _before_ spawning the child. 
+			   Thus, maximum number of connections is 
+		           conn_count[type]+1 */
 		conn = create_sql_connection(type);
 	}
 	conn->qid = qid;
@@ -749,13 +820,14 @@ attach_sql_connection(type, qid)
 /* Unattach the SQL connection. Move it to the end of the list.
  */
 void
-unattach_sql_connection(type, qid)
+unattach_sql_connection(type, qid, keepopen)
 	int    type;
 	qid_t  qid;
+	int keepopen;
 {
 	struct sql_connection *conn, *prev;
 
-	if (sql_cfg.keepopen == 0) {
+	if (keepopen == 0) {
 		free_sql_connection(type, qid);
 		return;
 	}
@@ -845,25 +917,56 @@ close_sql_connections(type)
 /* ************************************************************************* */
 
 int
-rad_sql_setup(type, qid)
+rad_sql_setup(type, req)
 	int type;
-	qid_t qid;
+	RADIUS_REQ *req;
 {
+	int dummy;
 	insist(type >= 0 && type < SQL_NSERVICE);
 
 	if (!sql_cfg.keepopen)
 		return 0;
 	
+	req->qid = rad_sql_get_qid();
 	if (sql_cfg.active[type]) {
-		attach_sql_connection(type, qid); 
+		attach_sql_connection(type, req->qid, &dummy); 
 	}
 	return 0;
 }
 
 void
-rad_sql_cleanup(type, qid)
+rad_sql_idle_check(void)
+{
+  	struct sql_connection *conn = NULL, *prev, *next;
+  	time_t now = time(NULL);
+
+  	prev = NULL;
+	
+  	if (sql_cfg.keepopen) {
+    		conn = conn_first;
+    		while (conn) {
+			next = conn->next;
+
+			if (master_process() &&
+			    conn->qid == 0 &&
+			    (now - conn->last_used) >= sql_cfg.idle_timeout) {
+				/* Close the idle connection */
+				debug(1,
+				      ("connection reached idle timeout: %p,%d",
+				       conn, conn->type));
+				close_sql_connection(conn, prev);
+			} else
+				prev = conn;
+
+			conn = next;
+		}
+	}
+}
+
+void
+rad_sql_cleanup(type, req)
 	int type;
-	qid_t qid;
+	RADIUS_REQ *req;
 {
 	insist(type >= 0 && type < SQL_NSERVICE);
 
@@ -871,7 +974,7 @@ rad_sql_cleanup(type, qid)
 		return;
 
 	if (sql_cfg.active[type])
-		unattach_sql_connection(type, qid);
+		unattach_sql_connection(type, req->qid, sql_cfg.keepopen);
 }
 
 void
@@ -921,53 +1024,55 @@ rad_sql_check_connect(type)
  * Perform normal accounting
  */ 
 void
-rad_sql_acct(authreq)
-	AUTH_REQ *authreq;
+rad_sql_acct(radreq)
+	RADIUS_REQ *radreq;
 {
 	int rc, count;
 	int status;
 	VALUE_PAIR *pair;
 	char *query;
 	struct sql_connection *conn;
+	int keepopen;
 	
 	if (!sql_cfg.doacct)
 		return;
 
-	if ((pair = pairfind(authreq->request, DA_ACCT_STATUS_TYPE)) == NULL) {
+	if ((pair = avl_find(radreq->request, DA_ACCT_STATUS_TYPE)) == NULL) {
 		/* should never happen!! */
 		radlog(L_ERR, _("no Acct-Status-Type record in rad_sql_acct()"));
 		return ;
 	}
 	status = pair->lvalue;
 
-	conn = attach_sql_connection(SQL_ACCT, (qid_t)authreq);
+        debug(1, ("qid %d", radreq->qid));
+	conn = attach_sql_connection(SQL_ACCT, radreq->qid, &keepopen);
 
 	switch (status) {
 	case DV_ACCT_STATUS_TYPE_START:
 		if (!sql_cfg.acct_start_query)
 			break;
-		query = radius_xlate(sql_cfg.buf.ptr, sql_cfg.buf.size,
+		query = radius_xlate(&stack,
 				     sql_cfg.acct_start_query,
-				     authreq->request, NULL);
-		rc = rad_sql_query(conn, query, NULL);
+				     radreq, NULL);
+		rc = disp_sql_query(sql_cfg.interface, conn, query, NULL);
 		sqllog(rc, query);
 		break;
 		
 	case DV_ACCT_STATUS_TYPE_STOP:
 		if (!sql_cfg.acct_stop_query)
 			break;
-		query = radius_xlate(sql_cfg.buf.ptr, sql_cfg.buf.size,
+		query = radius_xlate(&stack,
 				     sql_cfg.acct_stop_query,
-				     authreq->request, NULL);
-		rc = rad_sql_query(conn, query, &count);
+				     radreq, NULL);
+		rc = disp_sql_query(sql_cfg.interface, conn, query, &count);
 		sqllog(rc, query);
 		if (rc == 0 && count != 1) {
 			char *name;
 			char *session_id;
 
-			pair = pairfind(authreq->request, DA_USER_NAME);
+			pair = avl_find(radreq->request, DA_USER_NAME);
 			name = pair ? pair->strvalue : _("unknown");
-			pair = pairfind(authreq->request, DA_ACCT_SESSION_ID);
+			pair = avl_find(radreq->request, DA_ACCT_SESSION_ID);
 			session_id = pair ? pair->strvalue : _("unknown");
 			radlog(L_WARN, 
 			       _("SQL %s (%s) %d rows changed"),
@@ -978,137 +1083,260 @@ rad_sql_acct(authreq)
 	case DV_ACCT_STATUS_TYPE_ACCOUNTING_ON:
 		if (!sql_cfg.acct_nasup_query)
 			break;
-		query = radius_xlate(sql_cfg.buf.ptr, sql_cfg.buf.size,
+		query = radius_xlate(&stack,
 				     sql_cfg.acct_nasup_query,
-				     authreq->request, NULL);
-		rc = rad_sql_query(conn, query, &count);
+				     radreq, NULL);
+		rc = disp_sql_query(sql_cfg.interface, conn, query, &count);
 		sqllog(rc, query);
 		if (rc == 0) {
 			radlog(L_INFO,
 			       _("SQL: %d records updated writing acct-on info for NAS %s"),
 			       count,
-			       nas_name2(authreq));
+			       nas_request_to_name(radreq));
 		}
 		break;
 
 	case DV_ACCT_STATUS_TYPE_ACCOUNTING_OFF:
 		if (!sql_cfg.acct_nasdown_query)
 			break;
-		query = radius_xlate(sql_cfg.buf.ptr, sql_cfg.buf.size,
+		query = radius_xlate(&stack,
 				     sql_cfg.acct_nasdown_query,
-				     authreq->request, NULL);
-		rc = rad_sql_query(conn, query, &count);
+				     radreq, NULL);
+		rc = disp_sql_query(sql_cfg.interface, conn, query, &count);
 		sqllog(rc, query);
 		if (rc == 0) {
 			radlog(L_INFO,
 			       _("SQL: %d records updated writing acct-off info for NAS %s"),
 			       count,
-			       nas_name2(authreq));
+			       nas_request_to_name(radreq));
 		}
 		break;
 
 	case DV_ACCT_STATUS_TYPE_ALIVE:
 		if (!sql_cfg.acct_keepalive_query)
 			break;
-		query = radius_xlate(sql_cfg.buf.ptr, sql_cfg.buf.size,
+		query = radius_xlate(&stack,
 				     sql_cfg.acct_keepalive_query,
-				     authreq->request, NULL);
-		rc = rad_sql_query(conn, query, &count);
+				     radreq, NULL);
+		rc = disp_sql_query(sql_cfg.interface, conn, query, &count);
 		sqllog(rc, query);
 		if (rc != 0) {
 			radlog(L_INFO,
 			       _("SQL: %d records updated writing keepalive info for NAS %s"),
 			       count,
-			       nas_name2(authreq));
+			       nas_request_to_name(radreq));
 		}
 		break;
 		
 	}
 
-	if (!sql_cfg.keepopen)
-		unattach_sql_connection(SQL_ACCT, (qid_t)authreq);
+	if (!keepopen)
+		unattach_sql_connection(SQL_ACCT, conn->qid, keepopen);
+
+	if (query)
+		obstack_free(&stack, query);
 }
 
 
-int
-rad_sql_pass(req, authdata, passwd)
-	AUTH_REQ *req;
+char *
+rad_sql_pass(req, authdata)
+	RADIUS_REQ *req;
 	char *authdata;
-	char *passwd;
 {
 	int   rc;
 	char *mysql_passwd;
 	struct sql_connection *conn;
+	char *query;
+	int keepopen;
 
 	if (sql_cfg.doauth == 0) {
 		radlog(L_ERR,
 		       _("SQL Auth specified in users file, but not in sqlserver file"));
-		return AUTH_FAIL;
+		return NULL;
 	}
 	
 	if (authdata) {
-		pairadd(&req->request,
-			create_pair(DA_AUTH_DATA,
-				    strlen(authdata),
-				    authdata, 0));
+		avl_add_pair(&req->request,
+			     avp_create(DA_AUTH_DATA,
+					strlen(authdata),
+					authdata, 0));
 	}
-	radius_xlate(sql_cfg.buf.ptr, sql_cfg.buf.size,
-		     sql_cfg.auth_query,
-		     req->request, NULL);
-	pairdelete(&req->request, DA_AUTH_DATA);
+	query = radius_xlate(&stack, sql_cfg.auth_query, req, NULL);
+	avl_delete(&req->request, DA_AUTH_DATA);
 	
-	conn = attach_sql_connection(SQL_AUTH, (qid_t)req);
-	mysql_passwd = rad_sql_getpwd(conn, sql_cfg.buf.ptr);
+        debug(1, ("qid %d", req->qid));
+	conn = attach_sql_connection(SQL_AUTH, req->qid, &keepopen);
+	mysql_passwd = disp_sql_getpwd(sql_cfg.interface, conn, query);
 	
-	if (!mysql_passwd) {
-		rc = AUTH_NOUSER;
-	} else {
+	if (mysql_passwd) 
 		chop(mysql_passwd);
-		if (strcmp(mysql_passwd, md5crypt(passwd, mysql_passwd)) == 0)
-			rc = AUTH_OK;
-		else
-			rc = AUTH_FAIL;
-		efree(mysql_passwd);
-	}
 	
-	if (!sql_cfg.keepopen)
-		unattach_sql_connection(SQL_AUTH, (qid_t)req);
+	if (!keepopen)
+		unattach_sql_connection(SQL_AUTH, conn->qid, keepopen);
 	
-	return rc;
+	if (!query)
+		obstack_free(&stack, query);
+	
+	return mysql_passwd;
 }
 
 int
 rad_sql_checkgroup(req, groupname)
-	AUTH_REQ *req;
+	RADIUS_REQ *req;
 	char *groupname;
 {
 	int   rc = -1;
 	struct sql_connection *conn;
 	void *data;
 	char *p;
+	char *query;
+	int keepopen;
 	
-	if (sql_cfg.group_query == NULL) 
+	if (sql_cfg.doauth == 0 || sql_cfg.group_query == NULL) 
 		return -1;
 
-	radius_xlate(sql_cfg.buf.ptr, sql_cfg.buf.size,
-		     sql_cfg.group_query,
-		     req->request, NULL);
+        debug(1, ("qid %d", req->qid));
+	conn = attach_sql_connection(SQL_AUTH, req->qid, &keepopen);
+	if (!conn)
+		return -1;
 
-	conn = attach_sql_connection(SQL_AUTH, (qid_t)req);
-	data = rad_sql_exec(conn, sql_cfg.buf.ptr);
-	while (rc != 0 && rad_sql_next_tuple(conn, data) == 0) {
-		if ((p = rad_sql_column(data,0)) == NULL)
+	query = radius_xlate(&stack, sql_cfg.group_query, req, NULL);
+
+	data = disp_sql_exec(sql_cfg.interface, conn, query);
+	while (rc != 0 && disp_sql_next_tuple(sql_cfg.interface, conn, data) == 0) {
+		if ((p = disp_sql_column(sql_cfg.interface, data, 0)) == NULL)
 			break;
 		chop(p);
 		if (strcmp(p, groupname) == 0)
 			rc = 0;
 	}
-	rad_sql_free(conn, data);
+	disp_sql_free(sql_cfg.interface, conn, data);
 	
-	if (!sql_cfg.keepopen)
-		unattach_sql_connection(SQL_AUTH, (qid_t)req);
+	if (!keepopen)
+		unattach_sql_connection(SQL_AUTH, conn->qid, keepopen);
 	
+	if (query)
+		obstack_free(&stack, query);
 	return rc;
 }
 
+static int rad_sql_retrieve_pairs(struct sql_connection *conn,
+				  char *query,
+				  VALUE_PAIR **return_pairs,
+				  int op_too);
+
+int
+rad_sql_retrieve_pairs(conn, query, return_pairs, op_too)
+	struct sql_connection *conn;
+	char *query;
+	VALUE_PAIR **return_pairs;
+	int op_too;
+{
+	void *data;
+	int i;
+	
+        data = disp_sql_exec(sql_cfg.interface, conn, query);
+	if (!data)
+		return 0;
+	
+	for (i = 0; disp_sql_next_tuple(sql_cfg.interface, conn, data) == 0;
+	     i++) {
+		VALUE_PAIR *pair;
+		char *attribute;
+		char *value;
+		int op;
+		
+		if (!(attribute = disp_sql_column(sql_cfg.interface, data, 0))
+		    || !(value =  disp_sql_column(sql_cfg.interface, data, 1)))
+			break;
+		if (op_too) {
+			char *opstr;
+			opstr = disp_sql_column(sql_cfg.interface, data, 2);
+			if (!opstr)
+				break;
+			chop(opstr);
+			op = str_to_op(opstr);
+			if (op == NUM_OPERATORS) {
+				radlog(L_NOTICE,
+				       _("SQL: invalid operator: %s"), opstr);
+				continue;
+			}
+		} else
+			op = OPERATOR_EQUAL;
+
+		chop(attribute);
+		chop(value);
+		
+		pair = install_pair(attribute, op, value);
+		
+		if (pair)
+			avl_add_list(return_pairs, pair);
+	}
+
+	disp_sql_free(sql_cfg.interface, conn, data);
+	return i;
+}
+
+int
+rad_sql_reply_attr_query(req, reply_pairs)
+        RADIUS_REQ *req;
+        VALUE_PAIR **reply_pairs;
+{
+        VALUE_PAIR *request_pairs = req->request;
+	VALUE_PAIR *pair;
+	struct sql_connection *conn;
+	qid_t qid;
+	char *query;
+	int rc;
+	int keepopen;
+	
+	if (sql_cfg.doauth == 0 || !sql_cfg.reply_attr_query)
+		return 0;
+	
+        debug(1, ("qid %d", req->qid));
+	conn = attach_sql_connection(SQL_AUTH, req->qid, &keepopen);
+	
+	query = radius_xlate(&stack, sql_cfg.reply_attr_query, req, NULL);
+
+	rc = rad_sql_retrieve_pairs(conn, query, reply_pairs, 0);
+	
+        if (!keepopen) 
+                unattach_sql_connection(SQL_AUTH, conn->qid, keepopen);
+
+	if (query)
+		obstack_free(&stack, query);
+	return rc == 0;
+}
+
+int
+rad_sql_check_attr_query(req, return_pairs)
+        RADIUS_REQ *req;
+        VALUE_PAIR **return_pairs;
+{
+        VALUE_PAIR *request_pairs = req->request;
+	VALUE_PAIR *pair;
+	struct sql_connection *conn;
+	qid_t qid;
+	char *query;
+	int rc;
+	int keepopen;
+	
+	if (sql_cfg.doauth == 0 || !sql_cfg.check_attr_query)
+		return 0;
+	
+        debug(1, ("qid %d", req->qid));
+	conn = attach_sql_connection(SQL_AUTH, req->qid, &keepopen);
+	
+	query = radius_xlate(&stack, sql_cfg.check_attr_query, req, NULL);
+
+	rc = rad_sql_retrieve_pairs(conn, query, return_pairs, 1);
+	
+        if (!keepopen) 
+                unattach_sql_connection(SQL_AUTH, conn->qid, keepopen);
+
+	if (query)
+		obstack_free(&stack, query);
+	return rc == 0;
+}
+
 #endif
