@@ -1,26 +1,24 @@
 /* This file is part of GNU RADIUS.
- * Copyright (C) 2000, Sergey Poznyakoff
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
- */
+   Copyright (C) 2000, Sergey Poznyakoff
+  
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+   
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+  
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation,
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-#define RADIUS_MODULE 14
+#define RADIUS_MODULE_STAT_C
 #ifndef lint
 static char rcsid[] = 
-"@(#) $Id: stat.c,v 1.9 2001/03/27 09:54:18 gray Exp $";
+"@(#) $Id: stat.c,v 1.16 2001/12/25 15:49:09 gray Exp $";
 #endif
 
 #ifdef HAVE_CONFIG_H
@@ -209,21 +207,23 @@ stat_update(ut, status)
 	PORT_STAT *port;
 	long dt;
 	
-	nas = nas_find(ntohl(ut->nas_address));
+	nas = nas_lookup_ip(ntohl(ut->nas_address));
 	if (!nas) {
 		radlog(L_WARN,
-		    _("stat_update(): portno %d: can't find nas for IP %I"),
+		    _("stat_update(): portno %d: can't find nas for IP %s"),
 		    ut->nas_port,
-		    ntohl(ut->nas_address));
+		    format_ipaddr(ntohl(ut->nas_address)));
 		return;
 	}
-
+	if (nas->ipaddr == 0) /* DEFAULT nas */
+		return;
+	
 	port = stat_find_port(nas, ut->nas_port);
 	if (!port) {
 		radlog(L_WARN,
-		    _("stat_update(): port %d not found on NAS %I"),
+		    _("stat_update(): port %d not found on NAS %s"),
 		    ut->nas_port,
-		    ntohl(ut->nas_address));
+		    format_ipaddr(ntohl(ut->nas_address)));
 		return;
 	}
 
@@ -291,11 +291,12 @@ void
 stat_count_ports()
 {
 	NAS *nas;
+	struct nas_stat *statp;
 	PORT_STAT *port;
-	extern NAS *naslist;
 	
-	for (nas = naslist; nas; nas = nas->next) {
-		nas->nas_stat->ports_active = nas->nas_stat->ports_idle = 0;
+	for (nas = nas_next(NULL); nas; nas = nas_next(nas)) {
+		statp = nas->app_data;
+		statp->ports_active = statp->ports_idle = 0;
 	}
 	
 	radstat.port_active_count = radstat.port_idle_count = 0;
@@ -304,16 +305,17 @@ stat_count_ports()
 		if (port->ip == 0)
 			break;
 
-		nas = nas_find(port->ip);
+		nas = nas_lookup_ip(port->ip);
 		if (!nas) {
 			/* Silently ignore */
 			continue;
 		}
+		statp = nas->app_data;
 		if (port->active) {
-			nas->nas_stat->ports_active++;
+			statp->ports_active++;
 			radstat.port_active_count++;
 		} else {
-			nas->nas_stat->ports_idle++;
+			statp->ports_idle++;
 			radstat.port_idle_count++;
 		}
 	}
@@ -327,6 +329,8 @@ findportbyindex(ind)
 	PORT_STAT *p;
 	int i;
 
+        if (ind < 1 || ind > STAT_MAX_PORT_COUNT)
+                return NULL;
 	for (i = 1, p = stat_base+1;
 	     i < ind && p < stat_base + STAT_MAX_PORT_COUNT + 1 && p->ip;
 	     i++, p++) /* empty */ ;
