@@ -1,23 +1,22 @@
 /* This file is part of GNU RADIUS.
- * Copyright (C) 2001, Sergey Poznyakoff
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
- */
+   Copyright (C) 2001, Sergey Poznyakoff
+  
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+  
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+  
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation, 
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+
 #ifndef lint
-static char rcsid[] = "@(#) $Id: timestr.c,v 1.8 2001/04/19 15:53:27 gray Exp $";
+static char rcsid[] = "@(#) $Id: timestr.c,v 1.10 2001/11/27 12:13:36 gray Exp $";
 #endif
 
 #ifdef HAVE_CONFIG_H
@@ -97,9 +96,9 @@ timespan_day(ts, day, start, stop)
 	 *        just return
 	 * b. R complements F
 	 *        Append R to F
-	 * c. R preceedes F
+	 * c. R precedes F
 	 *        insert R before F
-	 * d. F preceedes R
+	 * d. F precedes R
 	 *        continue
 	 */
 	prev = NULL;
@@ -319,14 +318,34 @@ int main(int argc, char **argv)
 	TIMESPAN       *ts;
 	char           *p;
 	unsigned       rest;
-	
 	int i;
-	if (argc != 2) {
-		fprintf(stderr, "Usage: %s timestring\n", argv[0]);
-		exit(1);
-	}
+	struct tm *tmp, tm;
 
 	time(&t);
+	tmp = localtime(&t);
+	tm = *tmp;
+	
+	switch (argc) {
+	default:
+		fprintf(stderr, "Usage: %s timestring [dow hh mm]\n",
+			argv[0]);
+		exit(1);
+	case 5:
+		tm.tm_min = atoi(argv[4]);
+	case 4:
+		tm.tm_hour = atoi(argv[3]);
+	case 3:
+		tm.tm_wday = 0;
+		tm.tm_mday += atoi(argv[2]);
+		tm.tm_yday += atoi(argv[2]);
+		t = mktime(&tm);
+		break;
+	case 2:
+		break;
+	}
+
+	printf("ctime: %s", ctime(&t));
+	
 	if (ts_parse(&ts, argv[1], &p)) {
 		printf("bad timestring near %s\n", p);
 		return 1;
