@@ -1,22 +1,21 @@
 /* This file is part of GNU RADIUS.
- * Copyright (C) 2000,2001, Sergey Poznyakoff
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
- */
-/* $Id: version.c,v 1.6 2001/04/19 15:52:56 gray Exp $ */
+   Copyright (C) 2000,2001, Sergey Poznyakoff
+  
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+   
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+  
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation,
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+
+/* $Id: version.c,v 1.16 2001/12/20 16:53:56 gray Exp $ */
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
@@ -45,6 +44,9 @@ static char *sys_def[] = {
 #if defined(__NetBSD__)
 	"NetBSD",
 #endif
+#if defined(__OpenBSD__)
+        "OpenBSD",
+#endif
 #if defined(sun)
 	"sun",
 #endif
@@ -77,28 +79,31 @@ static char *compile_flag_str[] = {
 #if defined(PWD_SHADOW)
 	"PWD_SHADOW",
 #endif
+#if defined(USE_SERVER_GUILE)
+	"USE_SERVER_GUILE",
+#endif
 #if defined(USE_PAM)
 	"USE_PAM",
 #endif
-#if defined(DBM)
-	"DBM",
-#endif
-#if defined(NDBM)
-	"NDBM",
-#endif
-#if defined(USE_SQL)	
-# if USE_SQL == SQL_MYSQL
-	"USE_SQL=SQL_MYSQL",
-# elif USE_SQL == SQL_POSTGRES
-	"USE_SQL=SQL_POSTGRES",
-# endif	
-#endif /* defined(USE_SQL) */
+#if defined(USE_DBM)
+# if USE_DBM == DBM_DBM	
+	"USE_DBM=DBM",
+# elif USE_DBM == DBM_NDBM
+	"USE_DBM=NDBM",
+# endif
+#endif /* USE_DBM */
+#ifdef USE_SQL_MYSQL
+	"USE_SQL_MYSQL",
+#endif	    
+#ifdef USE_SQL_POSTGRES
+	"USE_SQL_POSTGRES",
+#endif	
+#ifdef USE_SQL_ODBC
+	"USE_SQL_ODBC",
+#endif	
 #if defined(USE_SNMP)
 	"USE_SNMP",
 #endif
-#if defined(USE_NOTIFY)
-	"USE_NOTIFY",
-#endif
 #if defined(USE_LIVINGSTON_MENUS)
 	"USE_LIVINGSTON_MENUS",
 #endif
@@ -153,46 +158,47 @@ version()
 {
 	int i;
 	
-	fprintf(stderr, _("%s: GNU Radius version %s"), progname, VERSION);
+	printf(_("%s: GNU Radius version %s"), progname, VERSION);
 #ifdef BUILD_TARGET
-	fprintf(stderr, " (%s)", BUILD_TARGET);
+	printf(" (%s)", BUILD_TARGET);
 #endif
-	fprintf(stderr, "\n");
+	printf("\n");
 
-	fprintf(stderr, _("Compilation platform: "));
+	printf(_("Compilation platform: "));
 	for (i = 0; sys_def[i]; i++)
-		fprintf(stderr, "%s ", sys_def[i]);
+		printf("%s ", sys_def[i]);
 
-	fprintf(stderr, _("\nDebugging flags: "));
+	printf(_("\nDebugging flags: "));
 	for (i = 0; debug_flag_str[i]; i++) {
-		fprintf(stderr, "%s ", debug_flag_str[i]);
+		printf("%s ", debug_flag_str[i]);
 	}
 
-	fprintf(stderr, _("\nCompilation flags: "));
+	printf(_("\nCompilation flags: "));
 	for (i = 0; compile_flag_str[i]; i++) {
-		fprintf(stderr, "%s ", compile_flag_str[i]);
+		printf("%s ", compile_flag_str[i]);
 	}
-	fprintf(stderr, "\n");
-	fprintf(stderr, "Ports in use:\n");
-	fprintf(stderr, " AUTH: %d\n", PW_AUTH_UDP_PORT);
-	fprintf(stderr, " ACCT: %d\n", PW_ACCT_UDP_PORT);
+	printf("\n");
+	printf("Ports in use:\n");
+	printf(" AUTH: %d\n", DEF_AUTH_PORT);
+	printf(" ACCT: %d\n", DEF_ACCT_PORT);
 #if defined(USE_MYSQL)
-	fprintf(stderr, " SQL : %d\n", RAD_SQL_PORT);
+	printf(" SQL : %d\n", RAD_SQL_PORT);
 #endif
-	fprintf(stderr, "Paths:\n");
-	fprintf(stderr, _(" configuration directory: %s\n"), RADIUS_DIR);
-	fprintf(stderr, _(" logging directory:       %s\n"), RADLOG_DIR);
-	fprintf(stderr, _(" accounting directory:    %s\n"), RADACCT_DIR);
+	printf("Paths:\n");
+	printf(_(" configuration directory: %s\n"), RADIUS_DIR);
+	printf(_(" logging directory:       %s\n"), RADLOG_DIR);
+	printf(_(" accounting directory:    %s\n"), RADACCT_DIR);
 #if defined(DENY_SHELL)
-	fprintf(stderr, _(" deny shell:              %s\n"), DENY_SHELL);
+	printf(_(" deny shell:              %s\n"), DENY_SHELL);
 #endif
 
 #ifdef RADIUS_PID
-	fprintf(stderr, _(" pidfile:                 %s\n"), RADIUS_PID);
+	printf(_(" pidfile:                 %s\n"), RADIUS_PID);
 #else
-	fprintf(stderr, _(" no pidfile\n"));
+	printf(_(" no pidfile\n"));
 #endif
 
+	printf("\nReport bugs to <%s>\n", bug_report_address);
 	exit(0);
 }
 
