@@ -1,40 +1,40 @@
 /* This file is part of GNU RADIUS.
- * Copyright (C) 2000, Sergey Poznyakoff
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
- */
+   Copyright (C) 2000, Sergey Poznyakoff
+ 
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+ 
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+ 
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation, 
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+
 #ifndef lint
 static char rcsid[] = 
-"$Id: client.c,v 1.7 2001/03/15 12:33:50 gray Exp $";
+"$Id: client.c,v 1.19 2001/12/16 12:07:00 gray Exp $";
 #endif
 
 #ifdef HAVE_CONFIG_H
 # include <config.h>
 #endif
 
+#include <stdio.h>
+#include <stdlib.h>
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
-#include <stdlib.h>
 #include <unistd.h>
 #include <errno.h>
 #include <ctype.h>
 
-#include <radiusd.h>
+#include <radius.h>
 #include <radclient.h>
 #include <slist.h>
 
@@ -42,29 +42,29 @@ int radclient_debug;
 
 static int radclient_build_request(RADCLIENT *config, SERVER *server, int code,
 				   VALUE_PAIR *pair);
-static AUTH_REQ * radclient_recv(UINT4 host, u_short udp_port,
+static RADIUS_REQ * radclient_recv(UINT4 host, u_short udp_port,
 				 char *secret, char *vector,
 				 char *buffer,
 				 int length);
-static AUTH_REQ * decode_buffer(UINT4 host, u_short udp_port, char *buffer,
+static RADIUS_REQ * decode_buffer(UINT4 host, u_short udp_port, char *buffer,
 				int length);
 static void random_vector(char *vector);
 static char * auth_code_str(int code);
 
 static struct keyword auth_codes[] = {
 #define D(a) #a, a	
-	D(PW_AUTHENTICATION_REQUEST),
-	D(PW_AUTHENTICATION_ACK),
-	D(PW_AUTHENTICATION_REJECT),
-	D(PW_ACCOUNTING_REQUEST),
-	D(PW_ACCOUNTING_RESPONSE),
-	D(PW_ACCOUNTING_STATUS),
-	D(PW_PASSWORD_REQUEST),
-	D(PW_PASSWORD_ACK),
-	D(PW_PASSWORD_REJECT),
-	D(PW_ACCOUNTING_MESSAGE),
-	D(PW_ACCESS_CHALLENGE),
-	D(PW_ASCEND_TERMINATE_SESSION),
+	D(RT_AUTHENTICATION_REQUEST),
+	D(RT_AUTHENTICATION_ACK),
+	D(RT_AUTHENTICATION_REJECT),
+	D(RT_ACCOUNTING_REQUEST),
+	D(RT_ACCOUNTING_RESPONSE),
+	D(RT_ACCOUNTING_STATUS),
+	D(RT_PASSWORD_REQUEST),
+	D(RT_PASSWORD_ACK),
+	D(RT_PASSWORD_REJECT),
+	D(RT_ACCOUNTING_MESSAGE),
+	D(RT_ACCESS_CHALLENGE),
+	D(RT_ASCEND_TERMINATE_SESSION),
 	0
 #undef D	
 };
@@ -81,7 +81,7 @@ auth_code_str(code)
 	return NULL;
 }
 
-AUTH_REQ *
+RADIUS_REQ *
 radclient_send(config, port_type, code, pair)
 	RADCLIENT *config;
 	int port_type;
@@ -99,7 +99,7 @@ radclient_send(config, port_type, code, pair)
 	struct timeval tm;
 	int result;
 	int i;
-	AUTH_REQ *req = NULL;
+	RADIUS_REQ *req = NULL;
 	SERVER *server;
 
 	if (port_type < 0 || port_type > 2) {
@@ -135,6 +135,12 @@ radclient_send(config, port_type, code, pair)
 		if (server->port[port_type] <= 0)
 			continue;
 		
+		if (radclient_debug) {
+			printf("server %s:%d\n",
+			       format_ipaddr(server->addr),
+			       server->port[port_type]);
+		}
+		
 		total_length = radclient_build_request(config, server,
 						       code, pair);
 	
@@ -150,10 +156,16 @@ radclient_send(config, port_type, code, pair)
 		sin->sin_family = AF_INET;
 		sin->sin_addr.s_addr = htonl(server->addr);
 		sin->sin_port = htons(server->port[port_type]);
-		
+
 		for (i = 0; i < config->retries; i++) {
-			sendto(sockfd, config->data_buffer, total_length, 0,
-			       &saremote, sizeof(struct sockaddr_in));
+			if (sendto(sockfd, config->data_buffer,
+				   total_length, 0,
+				   &saremote,
+				   sizeof(struct sockaddr_in)) == -1) {
+				radlog(L_ERR,
+				       "sendto: %s", strerror(errno));
+				break;
+			}
 
 			salen = sizeof (saremote);
 
@@ -191,6 +203,9 @@ radclient_send(config, port_type, code, pair)
 				break;
 			}
 		}
+
+		if (radclient_debug && !req)
+			printf("no reply\n");
 	} while (!req && (server = server->next) != NULL);
 	
 	close(sockfd);
@@ -207,11 +222,7 @@ radclient_build_request(config, server, code, pair)
 	int      total_length;
 	int      attrlen;
 	AUTH_HDR *auth;
-	char     passbuf[AUTH_PASS_LEN];
-	u_char   md5buf[256];
 	char     *ptr, *length_ptr;
-	int      secretlen;
-	int      i;
 	long     lval;
 	int      vendorcode, vendorpec;
 	
@@ -240,7 +251,8 @@ radclient_build_request(config, server, code, pair)
 	for (; pair; pair = pair->next) {
 
 		if (radclient_debug) {
-			radfprintf(stdout, "%10.10s: %A\n", _("send"), pair);
+			fprintf(stdout, "%10.10s: %s\n", 
+				   _("send"), format_pair(pair));
 		}
 
 		/*
@@ -269,32 +281,26 @@ radclient_build_request(config, server, code, pair)
 		*ptr++ = (pair->attribute & 0xFF);
 
 		switch (pair->type) {
-		case PW_TYPE_STRING:
+		case TYPE_STRING:
+			if (pair->strlength >= AUTH_STRING_LEN) {
+				radlog(L_ERR,
+  "radclient_build_request(): Attribute %d string value too long (%d bytes)",
+				       pair->attribute,
+				       pair->strlength);
+				goto overflow;
+			}
 			/* attrlen always < AUTH_STRING_LEN */
 			if (pair->attribute == DA_PASSWORD) {
-				char *p;
-
-				attrlen = AUTH_PASS_LEN;
-				CHECKSIZE(AUTH_PASS_LEN+2);
+				VALUE_PAIR *ppair;
+				ppair = avp_alloc();
+				encrypt_password(ppair, pair->strvalue,
+						 auth->vector, server->secret);
+				
+				attrlen = ppair->strlength;
+				CHECKSIZE(attrlen+2);
 				*ptr++ = attrlen + 2;
-				memset(passbuf, 0, attrlen);
-				memcpy(passbuf, pair->strvalue,
-				       pair->strlength < AUTH_PASS_LEN ?
-                                              pair->strlength : AUTH_PASS_LEN);
-
-				/* Calculate the MD5 Digest */
-				secretlen = strlen(server->secret);
-				strcpy(md5buf, server->secret);
-				memcpy(md5buf + secretlen, auth->vector,
-				       AUTH_VECTOR_LEN);
-				md5_calc(ptr, md5buf,
-					 secretlen + AUTH_VECTOR_LEN);
-				
-				/* Xor the password into the MD5 digest */
-				p = ptr;
-				for (i = 0; i < AUTH_PASS_LEN; i++) 
-					*p++ ^= passbuf[i];
-				
+				memcpy(ptr, ppair->strvalue, attrlen);
+				avp_free(ppair);
 			} else {
 				attrlen = pair->strlength;
 				CHECKSIZE(attrlen+2);
@@ -302,8 +308,8 @@ radclient_build_request(config, server, code, pair)
 				memcpy(ptr, pair->strvalue, attrlen);
 			}
 			break;
-		case PW_TYPE_INTEGER:
-		case PW_TYPE_IPADDR:
+		case TYPE_INTEGER:
+		case TYPE_IPADDR:
 			attrlen = sizeof(UINT4);
 			CHECKSIZE(attrlen+2);
 			*ptr++ = attrlen + 2;
@@ -329,7 +335,7 @@ overflow:
 }
 
 
-AUTH_REQ *
+RADIUS_REQ *
 radclient_recv(host, udp_port, secret, vector, buffer, length)
 	UINT4 host;
 	u_short udp_port;
@@ -369,7 +375,7 @@ radclient_recv(host, udp_port, secret, vector, buffer, length)
 }
 
 
-AUTH_REQ *
+RADIUS_REQ *
 decode_buffer(host, udp_port, buffer, length)
 	UINT4 host;
 	u_short udp_port;
@@ -385,24 +391,24 @@ decode_buffer(host, udp_port, buffer, length)
 	VALUE_PAIR	*first_pair;
 	VALUE_PAIR	*prev;
 	VALUE_PAIR	*pair;
-	AUTH_REQ	*authreq;
+	RADIUS_REQ	*radreq;
 
 	/*
 	 *	Pre-allocate the new request data structure
 	 */
 
-	authreq = alloc_request();
+	radreq = radreq_alloc();
 
 	auth = (AUTH_HDR *)buffer;
 
 	/*
 	 *	Fill header fields
 	 */
-	authreq->ipaddr = host;
-	authreq->udp_port = udp_port;
-	authreq->id = auth->id;
-	authreq->code = auth->code;
-	memcpy(authreq->vector, auth->vector, AUTH_VECTOR_LEN);
+	radreq->ipaddr = host;
+	radreq->udp_port = udp_port;
+	radreq->id = auth->id;
+	radreq->code = auth->code;
+	memcpy(radreq->vector, auth->vector, AUTH_VECTOR_LEN);
 
 	/*
 	 *	Extract attribute-value pairs
@@ -438,7 +444,7 @@ decode_buffer(host, udp_port, buffer, length)
 			       _("attribute %d too long, %d > %d"), attribute,
 			       attrlen, AUTH_STRING_LEN);
 		} else {
-			pair = alloc_pair();
+			pair = avp_alloc();
 			
 			pair->name = attr->name;
 			pair->attribute = attr->value;
@@ -447,7 +453,7 @@ decode_buffer(host, udp_port, buffer, length)
 
 			switch (attr->type) {
 
-			case PW_TYPE_STRING:
+			case TYPE_STRING:
 				pair->strvalue = alloc_string(attrlen + 1);
 				memcpy(pair->strvalue, ptr, attrlen);
 				pair->strvalue[attrlen] = '\0';
@@ -460,8 +466,8 @@ decode_buffer(host, udp_port, buffer, length)
 				prev = pair;
 				break;
 			
-			case PW_TYPE_INTEGER:
-			case PW_TYPE_IPADDR:
+			case TYPE_INTEGER:
+			case TYPE_IPADDR:
 				memcpy(&lval, ptr, sizeof(UINT4));
 				pair->lvalue = ntohl(lval);
 				if (first_pair == (VALUE_PAIR *)NULL) {
@@ -476,21 +482,21 @@ decode_buffer(host, udp_port, buffer, length)
 				radlog(L_ERR,
 				       _("    %s (Unknown Type %d)"),
 				       attr->name, attr->type);
-				free_pair(pair);
+				avp_free(pair);
 				break;
 			}
 
 			if (radclient_debug && pair) {
-				radfprintf(stdout, "%10.10s: %A\n", 
-						_("recv"), pair);
+				fprintf(stdout, "%10.10s: %s\n", 
+						_("recv"), format_pair(pair));
 			}
 
 		}
 		ptr += attrlen;
 		length -= attrlen + 2;
 	}
-	authreq->request = first_pair;
-	return authreq;
+	radreq->request = first_pair;
+	return radreq;
 }
 
 /*
@@ -515,13 +521,111 @@ random_vector(vector)
 /* ************************************************************************* */
 /* Initialization. */
 
+#define TOK_INVALID    0
+#define TOK_SOURCE_IP  1
+#define TOK_SERVER     2
+#define TOK_TIMEOUT    3
+#define TOK_RETRY      4
+
+static struct keyword kwd[] = {
+	"source_ip", TOK_SOURCE_IP,
+	"source-ip", TOK_SOURCE_IP,
+	"server", TOK_SERVER,
+	"timeout", TOK_TIMEOUT,
+	"retry", TOK_RETRY,
+	NULL
+};
+
+static int
+parse_client_config(client, argc, argv, file, lineno)
+	RADCLIENT *client;
+	int argc;
+	char **argv;
+	char *file;
+	int lineno;
+{
+	char *p;
+	SERVER serv;
+	
+	switch (xlat_keyword(kwd, argv[0], TOK_INVALID)) {
+	case TOK_INVALID:
+		radlog(L_ERR, "%s:%d: unknown keyword", file, lineno);
+		break;
+		
+	case TOK_SOURCE_IP:
+		client->source_ip = get_ipaddr(argv[1]);
+		break;
+		
+	case TOK_SERVER:
+		if (argc != 6) {
+			radlog(L_ERR, "%s:%d: wrong number of fields",
+			       file, lineno);
+			break;
+		}
+		memset(&serv, 0, sizeof serv);
+
+		serv.name = argv[1];
+		serv.addr = get_ipaddr(argv[2]);
+		if (!serv.addr) {
+			radlog(L_ERR,
+			       "%s:%d: bad IP address or host name",
+			       file, lineno);
+			break;
+		}
+		
+		serv.secret = argv[3];
+
+		serv.port[0] = strtol(argv[4], &p, 0);
+		if (*p) {
+			radlog(L_ERR,
+			       "%s:%d: bad port number %s",
+			       file, lineno, argv[4]);
+			break;
+		}
+
+		serv.port[1] = strtol(argv[5], &p, 0);
+		if (*p) {
+			radlog(L_ERR,
+			       "%s:%d: bad port number %s",
+			       file, lineno, argv[4]);
+			break;
+		}
+		
+		client->first_server =
+			radclient_append_server(client->first_server,
+						radclient_alloc_server(&serv));
+		break;
+		
+	case TOK_TIMEOUT:
+		client->timeout = strtol(argv[1], &p, 0);
+		if (*p) {
+			radlog(L_ERR,
+			       "%s:%d: bad timeout value", file, lineno);
+		}
+		break;
+		
+	case TOK_RETRY:
+		client->retries = strtol(argv[1], &p, 0);
+		if (*p) {
+			radlog(L_ERR,
+			       "%s:%d: bad retry value", file, lineno);
+		}
+		break;
+	}
+	return 0;
+}
+
+
 RADCLIENT *
-radclient_alloc(source_ip, bufsize)
+radclient_alloc(read_cfg, source_ip, bufsize)
+	int read_cfg;
 	UINT4 source_ip;
 	size_t bufsize;
 {
 	RADCLIENT *client;
-
+	struct timeval tv;
+	char *filename;
+	
 	client = emalloc(sizeof *client);
 
 	/* Provide default values */
@@ -531,8 +635,16 @@ radclient_alloc(source_ip, bufsize)
 	client->bufsize = bufsize ? bufsize : 4096;
 	client->first_server = NULL;
 	client->data_buffer = emalloc(client->bufsize);
-	client->messg_id = getpid() % 256;
 
+	gettimeofday(&tv, NULL);
+	srand(tv.tv_usec);
+	client->messg_id = random() % 256;
+
+	if (read_cfg) {
+		filename = mkfilename(radius_dir, "client.conf");
+		read_raddb_file(filename, 1, parse_client_config, client);
+		efree(filename);
+	}
 	return client;
 }
 
@@ -547,9 +659,7 @@ radclient_alloc_server(src)
 	server->addr = src->addr;
 	server->port[0] = src->port[0];
 	server->port[1] = src->port[1];
-	server->port[2] = src->port[2];
-	strncpy(server->secret, src->secret, AUTH_PASS_LEN);
-	server->secret[AUTH_PASS_LEN] = 0;
+	server->secret = make_string(src->secret);
 	return server;
 }
 
@@ -564,8 +674,7 @@ radclient_dup_server(src)
 	dest->name = dup_string(src->name);
 	dest->port[0] = src->port[0];
 	dest->port[1] = src->port[1];
-	dest->port[2] = src->port[2];
-	strncpy(dest->secret, src->secret, AUTH_PASS_LEN);
+	dest->secret = dup_string(src->secret);
 	return dest;
 }
 
@@ -578,6 +687,7 @@ radclient_free_server(server)
 	SERVER *server;
 {
 	free_string(server->name);
+	free_string(server->secret);
 	free_entry(server);
 }
 
