@@ -1,25 +1,23 @@
 /* This file is part of GNU RADIUS.
- * Copyright (C) 2000,2001, Sergey Poznyakoff
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- *
- */
+   Copyright (C) 2000,2001, Sergey Poznyakoff
+  
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU General Public License as published by
+   the Free Software Foundation; either version 2 of the License, or
+   (at your option) any later version.
+  
+   This program is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+   GNU General Public License for more details.
+  
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software Foundation, 
+   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 
 #ifndef lint
 static char rcsid[] = 
-"$Id: radwho.c,v 1.5 2001/03/12 11:49:38 gray Exp $";
+"$Id: radwho.c,v 1.11 2001/11/27 12:13:36 gray Exp $";
 #endif
 
 #if defined(HAVE_CONFIG_H)
@@ -34,13 +32,12 @@ static char rcsid[] =
 #include <sys/utsname.h>
 #include <ctype.h>
 #include <netinet/in.h>
-#if defined(HAVE_GETOPT_LONG)
-# include <getopt.h>
-#endif
 
+#include <getopt1.h>
 #include <sysdep.h>
 #include <radutmp.h>
-#include <radiusd.h>
+#include <radius.h>
+#include <radpaths.h>
 
 #define ALIGN_LEFT    0
 #define ALIGN_RIGHT   1
@@ -200,7 +197,7 @@ FORMAT realm_fmt = {
 FORMAT radwho_fmt[MAX_FMT];
 
 #define OPTSTR "Acd:D:e:f:FhHiI:lLno:su"
-#ifdef HAVE_GETOPT_LONG
+
 struct option longopt[] = {       
 	"all",               no_argument, 0, 'A',
 	"calling-id",        no_argument, 0, 'c',
@@ -221,10 +218,6 @@ struct option longopt[] = {
 	"secure",            no_argument, 0, 's',
 	0
 };
-#else
-# define longopt 0
-# define getopt_long(ac,av,os,lo,li) getopt(ac,av,os)
-#endif
 
 #define SIP_SMART    0
 #define SIP_NODOMAIN 1
@@ -497,7 +490,7 @@ local_who()
 			rt.type = P_CONSOLE;
 			strncpy(rt.login, ut.ut_name, RUT_NAMESIZE);
 			strncpy(rt.orig_login, ut.ut_host, RUT_NAMESIZE);
-#ifdef __svr4__
+#if defined __svr4__ || defined __sgi
 			rt.time = ut.ut_xtime;
 #else
 			rt.time = ut.ut_time;
@@ -513,20 +506,21 @@ local_who()
 void
 radius_who()
 {
-	FILE          *fp;
-	struct radutmp rt;
+	radut_file_t file;
+	struct radutmp *up;
 	print_header();
 
 	/*
 	 *	Show the users logged in on the terminal server(s).
 	 */
-	if ((fp = fopen(filename, "r")) == NULL)
+	if ((file = rut_setent(filename, 0)) == NULL)
 		return ;
 
-	while(fread(&rt, sizeof(rt), 1, fp) == 1) {
-		if (want_rad_record(&rt)) 
-			format_radutmp_entry(radwho_fmt, &rt);
+	while (up = rut_getent(file)) {
+		if (want_rad_record(up)) 
+			format_radutmp_entry(radwho_fmt, up);
 	}
+	rut_endent(file);
 }
 
 void
@@ -1113,7 +1107,6 @@ realm_name(ip)
  */
 
 char usage_str[] =
-#ifdef HAVE_GETOPT_LONG
 "usage: radwho [options] username\n"
 "Options are:\n"
 "       -A, --all            Print all entries, not only active ones.\n" 
@@ -1137,36 +1130,12 @@ char usage_str[] =
 "       -s, --secure         Secure mode: requires that the username be\n"
 "                            specified.\n"
 "       -u, --local-also     Display also local users.\n";
-#else
-"usage: radwho [options] username\n"
-"Options are:\n"
-"       -A                   Print all entries, not only active ones.\n" 
-"       -c                   Display CLID in second column.\n"
-"       -D {short|abbr|full}\n"
-"                            Change date representation format.\n"
-"       -d DIR               Specify Radius configuration directory.\n"
-"       -e STRING            Print STRING instead of an empty column.\n"
-"       -F                   Act as a finger daemon.\n"
-"       -f FILE              Use FILE instead of /var/log/radwtmp\n"
-"       -H                   Do not display header line.\n"
-"       -h                   Display this help.\n"
-"       -i                   Display session ID in the second column.\n"
-"       -I {smart|ip|nodomain}\n"
-"                            Change IP address representation format.\n"
-"       -L                   Display GNU license and exit.\n"
-"       -l                   Long output. All fields will be printed.\n"
-"                            Implies -D full -I smart.\n"
-"       -n                   Do not resolve hostnames. The same as -I ip.\n"
-"       -o                   Specify format line.\n"
-"       -s                   Secure mode: requires that the username be\n"
-"                            specified.\n"
-"       -u                   Display also local users.\n";
-#endif
 
 void
 usage()
 {
 	printf("%s", usage_str);
+	printf("\nReport bugs to <%s>\n", bug_report_address);
 }
 
 char license_text[] =
